import pandas as pd

# -------------------------------------------------
# Test to_csv date_format
# -------------------------------------------------
# Create DataFrame
a = pd.DataFrame()

# Create dates
a['dates'] = ['23/01/2015 19:37',
              '23/01/2015 20:08']

# Format dates
a.dates = pd.to_datetime(a.dates)

# Save
print("\nDF:")
print(a)
print("Saving...")
#a.to_csv('test-v0.1.csv')
#a.to_csv('test-v0.2.csv', date_format='%Y-%m-%d %H:%M:%S')

# -------------------------------------------------
# Test cleaning / replacing
# -------------------------------------------------
# Create DataFrame
a = pd.DataFrame()

# Create regexpmap
REGEX_MAP = {
    '\([^)]*\)': '',  # Remove everything between ().
    'species': '',  # Rename species for next regexp
    'sp(\.)?(\s|$)+': ' ',  # Remove sp from word.
    'strep(\.|\s|$)': 'streptococcus ',  # Complete
    'staph(\.|\s|$)': 'staphylococcus ',  # Complete
    '\s+': ' '  # Remove duplicated spaces.
}

# Create data
a['spaces'] = [' in   between ', ' sides ', 'end ', ' start', None]
a['species'] = [' sp.', ' sp', 'sp ', ' sp. ', 'species']
a['occus'] = ['haemolytic streptococcus',
              'haemolytic strep',
              'haemolytic strep.',
              'haemolytic strep. aureus',
              'haemolytic strep aureus']

a['occus2'] = ['strep.aureus',
               'staph.aureus',
               'methicillin resistant staph.aureus',
               'feo strepococcus',
               'streptococcus feo']

# Cleaned
cleaned = a.copy(deep=True)
cleaned = cleaned.replace(regex=REGEX_MAP)
cleaned = cleaned.apply(lambda x: x.str.strip() \
    if x.dtype == "object" else x)

# Show
print("-"*80)
print("\nOriginal")
print(a)
print("\nCleaned")
print(cleaned)

# ---------------------------------------------------
# Haemolytic
# --------------------------------------------------
# .. note: https://regex101.com/r/KFXCCM/1

# ----------------------------------------------------
# Test genus at the beginning
# ----------------------------------------------------
# Import regular expressions
import re

# Import function
from pyamr.datasets.clean import word_to_start

# Species examples
series = pd.Series(['is viridians, enteroccocus en',
                    'is viridians, enteroccocus',
                    'viridians enteroccocus',
                    'enterococcus viridians',
                    'non haemolytic feo enteroccocus',
                    'non-haemolytic enteroccocus',
                    'non haemolytic enteroccocus',
                    'vancomycin resistant enteroccocus',
                    None])


# Corrected
corrected = series.apply(word_to_start, w='enteroccocus')

# Show
print("-"*80)
print("\n\n\nRaw")
print(series)
print("\nCorrected")
print(corrected)

# ----------------------------------------------------
# Test hyphen
# ----------------------------------------------------
# Import
from pyamr.datasets.clean import hyphen_before

# Haemolytic examples
series= pd.Series(['non haemolytic something',
                   ' non haemolytic something',
                   ' when beta haemolytic something',
                   ' whn gamma haemolytic',
                   'non    haemolytic'])

# Correct it
corrected = series.apply(hyphen_before, w='haemolytic')

# Show
print("-"*80)
print("\n\n\nRaw:")
print(series)
print("\nCorrected:")
print(corrected)

# ----------------------------------------------------
# Full test
# ----------------------------------------------------
# In this section, we do a full test, specially for
# those examples that have shown to be problematic in
# the final microorganisms.csv outcome.
# Import clean common
from pyamr.datasets.clean import clean_common
from pyamr.datasets.registries import _clean_microorganism

# Create dataframe
df = pd.DataFrame()

# Add microorganism names.
df['microorganism_name'] = [
    'non haemolytic streptococcus aureus',
    'this non haemolytic streptococcus',
    'beta-haemolytic streptococcus group a',
    'beta-haemolytic streptococcus group b',
    'beta-haemolytic streptococcus group c',
    'beta-haemolytic streptococcus group c/g',
    'beta-haemolytic streptococcus group g',
    'this is haemolyticus',
    'perestreptococcus',
    'Coagulase negative staphylococcus',
    'Methicillin Resistant Staph.aureus',
    'mixed streptococcus alpha-haemolytic',
    'non-coliform lactose fermenting',
    'Non-haemolytic streptococcus',
    'Non-lactose Fermenting Coliform',
    'Vancomycin Resistant Enterococcus',
    ' non- lactose fermenting coliform',
    'non-lactose fermenting coliform',
    'mixed lactose fermenting coliform',
    'paenibacillus sp',
    'paenibacillus sp.',
    'paenibacillus sp..',
    'escherichia coli o157',
    '*** mrsa *** isolated',
    'streptococcus milleri group',
    'aspergillus fumigatus'
]

# Any alterations
df.microorganism_name = df.microorganism_name.str.upper()

# Clean
#aux = clean_common(df.copy(deep=True))
aux = _clean_microorganism(df.microorganism_name)

# Show
print("-"*80)
print("\nData:")
print(df)
print("\nCorrected:")
print(aux)


################################################################################
# Author:
# Date:
# Description:
#
# Copyright:
# 
################################################################################
from __future__ import division

# Libraries
import sys
import numpy as np
import pandas as pd


class Frequency(): # pragma: no cover
    """
    """

    # Attributes
    c_abx = 'ANTIBIOTIC'
    c_org = 'SPECIE'
    c_dat = 'DATE'
    c_out = 'SENSITIVITY'

    # -------------------------------------------------------------------------
    #
    # -------------------------------------------------------------------------
    def __init__(self, column_antibiotic=c_abx,
                 column_organism=c_org,
                 column_date=c_dat,
                 column_outcome=c_out,
                 column_labnumber=None,
                 dfmt='%Y-%M-%d'):
        """The constructor.

        Parameters
        ----------
        column_antibiotic : string
          The column name with the the antibiotic values

        column_organism : string
          The column name with the organism values

        column_date : string
          The column name with the dates

        column_labnumber : string
          The column name with the laboraory number

        Returns
        -------
        """
        # Create dictionary to rename columns
        self.rename_columns = {column_antibiotic: self.c_abx,
                               column_organism: self.c_org,
                               column_date: self.c_dat,
                               column_outcome: self.c_out}

    # -------------------------------------------------------------------------
    #
    # -------------------------------------------------------------------------
    def _by_category_groupby(self, by_category='pairs'):
        """This method returns the grouping list.

        Parameters
        ----------
        by_category : string
          The category to group by from pairs, organisms or antibiotics.

        Returns
        -------
        list
        """
        # Define how to return the overall results.
        if by_category == 'organisms':
            return [self.c_org, self.c_out]
        elif by_category == 'antibiotics':
            return [self.c_abx, self.c_out]
        elif by_category == 'pairs':
            return [self.c_org, self.c_abx, self.c_out]
        else:
            raise ValueError("The by_category parameter select must be one of"
                             "the following [pairs, organisms, antibiotics]; "
                             "the value <%s> was found." % by_category)

    def fit(self):
        """
        """
        pass

    def _compute_overall(self, dataframe, by_category='pairs'):
        """This method computes the overall frequency count.

        Parameters
        ----------
        dataframe : dataframe-like
          The dataframe with the microbiology data

        by_category

        Returns
        -------
        dataframe
        """
        # Get the definition of groupby
        groupby = self._by_category_groupby(by_category)
        # Compute results
        return dataframe.groupby(groupby).size().unstack().fillna(0)

    def _compute_independent(self, dataframe, by_category='pairs', fs='1D'):
        """This method computes the independent time intervals frequency.

        Parameters
        ----------
        dataframe:  dataframe-like
          The microbiology dataframe with the following columns.

        by_category: string
          The category to group the outcomes. The outcomes are grouped in pairs
          formed by (organism, antibiotic) by default. However, these can be
          also grouped by organisms or antibiotics.

        fs : string
          The frequency sample (e.g. 1D, 1M, 7D, ...)

        Returns
        -------
        dataframe
        """
        # Format the dataframe to have datetime index
        dataframe = dataframe.reset_index()
        dataframe = dataframe.set_index(self.c_dat)

        # Get the definition of groupby
        groupby = [pd.Grouper(freq=fs)] + self._by_category_groupby(by_category)

        # Compute independent window
        dataframe = dataframe.groupby(groupby).size().unstack().fillna(0)

        # Resample
        dataframe = dataframe.reset_index()
        dataframe = dataframe.set_index(self.c_dat) \
            .groupby([self.c_org, self.c_abx]) \
            .resample(fs).mean() \
            .fillna(0)

        # Return
        return dataframe

    def _compute_overlapping(self, dataframe, by_category='pairs',
                             wshift='1D',
                             wsize='2D'):
        """This method computes the overlapping time intervals frequency

        Parameters
        ----------
        dataframe:  dataframe-like
          The microbiology dataframe with the following columns.

        by_category: string
          The category to group the outcomes. The outcomes are grouped in pairs
          formed by (organism, antibiotic) by default. However, these can be
          also grouped by organisms or antibiotics.

        wshift : string
          The shift between consecutive windows (OTI).

        wsize : integer
          The size of the window (OTI)

        Returns
        -------
        dataframe
        """
        # Format the dataframe to have datetime index
        dataframe = self._compute_independent(dataframe=dataframe,
                                              by_category=by_category,
                                              fs=wshift)

        # Reset index
        dataframe = dataframe.reset_index()

        # Compute rolling window
        dataframe = dataframe.groupby([self.c_org, self.c_abx], as_index=False) \
            .apply(lambda x: x.set_index(self.c_dat) \
                   .rolling(window=wsize).sum()).fillna(0)

        # Drop index level (which appears when executing previous code)
        dataframe.index = dataframe.index.droplevel()


        print(dataframe)
        print(self.c_dat, self.c_org, self.c_abx)

        # Resample
        dataframe = dataframe.reset_index()
        dataframe = dataframe.set_index(self.c_dat) \
            .groupby([self.c_org, self.c_abx]) \
            .resample(wshift).mean() \
            .fillna(0)

        # Return
        return dataframe

    def compute(self, dataframe, strategy='overall',
                by_category='pairs',
                fs=None,
                wshift=None,
                wsize=None):
        """This function computes the frequencies.

        The method allows to compute the overall frequencies, the frequencies for
        independent time intervals (ITI) such as monthly or yearly and the
        frequencies for overlapping time intervals (OTI) in which the parameters
        wshift and wsize need to be specified.

        Parameters
        ----------
        dataframe:  dataframe-like
          The microbiology dataframe with the following columns.

        by_category: string
          The category to group the outcomes. The outcomes are grouped in pairs
          formed by (organism, antibiotic) by default. However, these can be
          also grouped by organisms or antibiotics.

        fs : string
          The frequency sample (e.g. 1D, 1M, 7D, ...)

        wshift : integer
          The shift between consecutive windows (OTI).

        wsize : integer
          The size of the window (OTI)

        Returns
        -------
        dataframe
        """
        # Check that it is a dataframe
        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError("The instance passed as argument needs to be a pandas "
                            "DataFrame. Instead, a <%s> was found. Please convert "
                            "the input accordingly." % type(dataframe))

        # Rename columns
        dataframe = dataframe.rename(columns=self.rename_columns, copy=True)

        # Ensure date columns has date objects.
        dataframe[self.c_dat] = pd.to_datetime(dataframe[self.c_dat])

        # -----------------------
        # Compute
        # -----------------------
        if strategy == 'overall':
            return self._compute_overall(dataframe=dataframe,
                                         by_category=by_category)

        elif strategy == 'ITI':
            return self._compute_independent(dataframe=dataframe,
                                             by_category=by_category,
                                             fs=fs)

        elif strategy == 'OTI':
            return self._compute_overlapping(dataframe=dataframe,
                                             by_category=by_category,
                                             wshift=wshift,
                                             wsize=wsize)


if __name__ == '__main__': # pragma: no cover

    """
    # Import libraries
    import sys
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    # Import own module
    sys.path.append('../../')

    # Import specific libraries
    from pyAMR.datasets import load

    # Set matplotlib
    mpl.rcParams['xtick.labelsize'] = 9
    mpl.rcParams['ytick.labelsize'] = 9
    mpl.rcParams['axes.titlesize'] = 11
    mpl.rcParams['legend.fontsize'] = 9

    # -----------------------
    # Load data
    # -----------------------
    # Load sample data
    data = load.dataset_epicimpoc_susceptibility_year(nrows=1000000)

    # Keep only relevant columns
    data = data[['antibioticCode',
                 'organismCode',
                 'dateReceived',
                 'sensitivity']]

    # Filter for two examples
    is_org = data['organismCode'] == 'ECOL'
    is_abx = data['antibioticCode'].isin(['ATAZ'])
    data = data[is_abx & is_org]

    # -------------------------
    # Create frequency instance
    # -------------------------
    # Create instance
    freq = Frequency(column_antibiotic='antibioticCode',
                     column_organism='organismCode',
                     column_date='dateReceived',
                     column_outcome='sensitivity')

    # ------------------------
    # Examples compute overall
    # ------------------------
    # Examples compute overall
    pairs = freq.compute(data, by_category='pairs')
    antibiotics = freq.compute(data, by_category='antibiotics')
    organisms = freq.compute(data, by_category='organisms')

    # Show
    # print pairs.head(10)
    # print antibiotics.head(10)
    # print organisms.head(10)

    # -------------------------------------------
    # Examples compute independent time intervals
    # -------------------------------------------
    # Examples compute ITI
    daily = freq.compute(data, strategy='ITI',
                         by_category='pairs',
                         fs='1D')

    monthly = freq.compute(data, strategy='ITI',
                           by_category='pairs',
                           fs='1M')

    # Show
    # print daily.head(10)
    # print monthly.head(10)

    # -------------------------------------------
    # Examples compute overlapping time intervals
    # -------------------------------------------
    # Examples compute OTI (daily)
    oti_1 = freq.compute(data, strategy='OTI',
                         by_category='pairs',
                         wshift='1D',
                         wsize=5)

    # Examples compute OTI (monthly)
    oti_2 = freq.compute(data, strategy='OTI',
                         by_category='pairs',
                         wshift='1M',
                         wsize=2)
    # ----------------
    # Plot
    # ----------------
    # Show comparison for each pair
    f, axes = plt.subplots(4, 1, figsize=(15, 8))

    # Flatten axes
    axes = axes.flatten()

    # Plot ITI (daily)
    for i, (pair, group) in enumerate(daily.groupby(level=[0, 1])):
        group.index = group.index.droplevel([0, 1])
        group.sum(axis=1).plot(marker='o', ms=3, label=pair,
                               linewidth=0.5, markeredgecolor='k', markeredgewidth=0.3,
                               ax=axes[0])

    # Plot ITI (monthly)
    for i, (pair, group) in enumerate(monthly.groupby(level=[0, 1])):
        group.index = group.index.droplevel([0, 1])
        group.sum(axis=1).plot(marker='o', ms=3, label=pair,
                               linewidth=0.5, markeredgecolor='k', markeredgewidth=0.3,
                               ax=axes[1])

    # Plot OTI
    for i, (pair, group) in enumerate(oti_1.groupby(level=[0, 1])):
        group.index = group.index.droplevel([0, 1])
        group.sum(axis=1).plot(marker='o', ms=3, label=pair,
                               linewidth=0.5, markeredgecolor='k', markeredgewidth=0.3,
                               ax=axes[2])

    # Plot OTI
    for i, (pair, group) in enumerate(oti_2.groupby(level=[0, 1])):
        group.index = group.index.droplevel([0, 1])
        group.sum(axis=1).plot(marker='o', ms=3, label=pair,
                               linewidth=0.5, markeredgecolor='k', markeredgewidth=0.3,
                               ax=axes[3])

    # Set legend
    for ax in axes:
        ax.legend()
        ax.set_xlabel('')
        ax.grid(True)

    # Set titles
    axes[0].set_ylabel('Daily')
    axes[1].set_ylabel('Monthly')
    axes[2].set_ylabel('OTI(1D,5)')
    axes[3].set_ylabel('OTI(1M,2)')

    # Show
    plt.show()
    """


###############################################################################
# Author: Bernard Hernandez
# Filename: generic.py
# Date: 02/09/2015
# Description: Methods to clean data.
#
###############################################################################

# Libraries.
from os import listdir
import numpy as np
import pandas as pd
import datetime
import glob
import json
import re

#-------------------------------------------------------------------------
#                         generic functions
#-------------------------------------------------------------------------
def _apply(df, cols, fun):
  """
  """
  if isinstance(cols, str): 
    cols = df.columns if cols=='all' else [cols]
  for col in cols:
    if col in df.columns:
      df[col] = df[col].apply(fun)
  return df

def _lower(x):
  """This method lowers the string"""
  try: 
    return x.lower()
  except: 
    return x

def _upper(x):
  """This method uppers the string"""
  try: 
    return x.upper()
  except: 
    return x

def _delete_spaces(x):
  """
  """
  try: 
    return ' '.join(x.split())
  except: 
    return x

def _delete_digits(x):
  """
  """
  try: 
    return ''.join([i for i in x if not i.isdigit()])
  except: 
    return x

def _delete_types(x):
  """
  """
  try:
    if not isinstance(x, str): return x
    if '(second type)' in x: return x.replace(' (second type)', '')
    if '(2nd type)' in x: return x.replace(' (2nd type)', '')
    if '(third type)' in x: return x.replace(' (third type)', '')
    if '(3rd type)' in x: return x.replace(' (3rd type)', '')
    return x
  except Exception as e:
    print(e)
    return x

def _delete_parenthesis_content(x):
  """
  """
  try:
    if not isinstance(x, str): return x
    if re.search(r'\([^)]*\)',x) is not None:
      x = re.sub(r'\([^)]*\)', '', x)
      return _delete_spaces(x)
    return x
  except Exception as e:
    return x

def _fix_sp_ending(x):
  """
  """
  try:
    if not isinstance(x, str): return x
    if x.endswith('sp'): return x+'.'
    if x.endswith('sp..'): return x[:-1]
    if ' sp ' in x: return x.replace(' sp ', ' sp. ')
    if 'species' in x: return x.replace('species','sp.')
    return x
  except Exception as e:
    print(e)
    return x

def _fix_sp_abbreviation(x):
  """
  """
  try:
    if not isinstance(x, str): return x
    if 'streptococci' in x: return x.replace('streptococci', 'streptococcus')
    if "strep." in x: return x.replace('strep.', 'streptococcus')
    if "staph." in x: return x.replace('staph.', 'staphylococcus ')
    return x
  except Exception as e:
    print(e)
    return x

#-------------------------------------------------------------------------
#                         GENERAL METHODS
#-------------------------------------------------------------------------
def lettercase(df, cols='all', funct=_lower):
  """
  """
  if isinstance(funct, str): 
    if funct=='lower': return _apply(df, cols, _lower)
    if funct=='upper': return _apply(df, cols, _upper)
  else:
    return _apply(df, cols, funct)

def delete_spaces(df, cols='all'):
  """
  """
  return _apply(df, cols, _delete_spaces)

def delete_parenthesis(df, cols='all'):
  """
  """
  return _apply(df, cols, _delete_parenthesis_content)

#-------------------------------------------------------------------------
#                      SENSITIVTY ORIENTED METHODS
#-------------------------------------------------------------------------
def delete_numbers(df, cols='all'):
  """
  """
  return _apply(df, cols, _delete_digits)

def delete_types(df, cols='all'):
  """
  """
  return _apply(df, cols, _delete_types)

def ending_specie(df, cols='all'):
  """
  """
  return _apply(df, cols, _fix_sp_ending)

def specie_abbreviation(df, cols='all'):
  """
  """
  return _apply(df, cols, _fix_sp_abbreviation)


# -----------------------------------------------------------
#
# -----------------------------------------------------------










def transform_rename(filepath, df):
  """This function rename codes (i.e. PAER1,PAER2 to PEAR)

  Parameters
  ----------

  Return
  ------
  """
  def transform_attribute(code_map, attr_name, df):
    """
    """
    for k in code_map:
      codes = code_map[k]
      is_in = df[attr_name].isin(codes)
      df.loc[is_in, attr_name] = k
    return df

  # Read config information.
  file = open(filepath) 
  config_info = json.load(file)
  print(config_info)
  # Variables.
  rename_drugs = config_info['drugs']['rename']
  rename_bugs = config_info['bugs']['rename']
  # Rename
  df = transform_attribute(rename_drugs, 'drugCode', df)
  df = transform_attribute(rename_bugs, 'organismCode', df)
  return df
  


###############################################################################
# Author: Bernard Hernandez
# Filename: acronym.py
# Date: 02/09/2015
# Description:
# 
# This script creates a map to solve possible issues of misspelled antibiotics,
# or differences as upper/lower cases, unnecessary spaces, missing codes, ...
# Hence, it creates a conversion between the original antibiotic name to an
# homogenous antibiotic name, and the original code and an homogeneous code.
#
###############################################################################

# Libraries
import pandas as pd

# --------------------------------------------------------------------------
#                              helper methods
# --------------------------------------------------------------------------
def acronym(name, length=3, prefix='A_'):
  """This method created the acronym for a given string

  Parameters
  ----------
  name: string-like
    The string to construct the acronym

  length: int-like
    The number of letters used for the first word of the acronym

  prefix: string-like
    The prefix to add to the acronym for identification purposes

  Returns
  -------
  string
  """
  # Check that the acronym is not nul
  # Remove spaces
  name = ' '.join(name.split())
  # Array with individual words
  words = name.split()   
  # Check there are words to create acronym
  if not len(words):
    return ('%sNONE' % (prefix)).upper()
  # Create acronym for other words
  acronym = [w[0] for w in words[1:]]
  # Create acronym adding first word
  acronym = '%s%s%s' % (prefix, words[0][:length], ''.join(acronym))
  # Return
  return acronym.upper()


def _check_acronym_conflicts(transdict):
  """This method check that all acronyms are unique.

  Parameters
  ----------
  transdict: dict-like
    The dictionar containing the string and the corresponding acronym.

  Returns
  -------
  """
  # Import library
  from collections import Counter
  # Count number of times the acronym appears
  counter = Counter(transdict.values())
  # Create a dict with the duplicated elements
  duplicated = {k:v for k,v in counter.items() if v>1}
  # Raise an error
  if len(duplicated):
    raise ValueError("The following acronyms within the dictionary "
                     "are not unique %s." % duplicated)


class AcronymBuilder:
  """
  """

  # Attributes
  _func = acronym

  # Constructor
  def __init__(self):
    """
    """
    pass
  
  def has_acronym():
    pass

  def has_value():
    pass

  def update(self):
    pass

  def fit(self, values, acronyms):
    """This method creates a dictionary with the acronyms

    TODO: Instead of just checking whether there are acronym conflicts or not
    ensure that always unique acronyms are created. Note that the implemented
    approach to create acronyms does not ensure they are always unique.

    Parameters
    ----------
    values: array-like (contains strings)
      The array with the possible values

    acronyms: array-like (contains strings)
      The array with the corresponding acronyms.

    Returns
    -------
    """
    # Check that lengths are the same
    if len(values)!=len(acronyms):
      raise ValueError("The length of the parameters values (%s) and "
                       "acronyms (%s) mismatch. They must be the same "
                       "length." % (len(values), len(acronyms)))

    # Remove duplicated spaces within values
    values = [' '.join(v.split()).lower() for v in values]

    # Create acronyms for nan entries
    acronyms = [acronym(v) if pd.isnull(a) else a 
      for v,a in zip(values, acronyms)]

    # Create translation dictionary
    self._transdict = dict(zip(values, acronyms))
    self._reverdict = dict(zip(acronyms, values))

    # Check acronym conflicts
    _check_acronym_conflicts(self._transdict)
    _check_acronym_conflicts(self._reverdict)

    # Return
    return self


  def transform(self, values):
    """This method returns the acronyms for given values

    Parameters
    ----------
    values: array-like

    Returns
    -------
    array with acronyms
    """
    return [self._transdict[value] for value in values]

  def inverse_transform(self, acronyms):
    """This method returns the values for given acronyms.

    Parameters
    ----------
    values: array-like

    Returns
    -------
    array with acronyms
    """
    return [self._reverdict[acronym] for acronym in acronyms]

  












if __name__ == '__main__':

  # Libraries
  import sys

  # Import specific libraries
  from pyamr.datasets import load


  # ---------------------
  # Example I
  # ---------------------
  # This example shows how to use the acronym builder to generate acronyms
  # to be used for the EPiC IMPOC research. In particular, it creates the
  # acronyms for the organisms.

  # Load data
  data = load.dataset_epicimpoc_susceptibility_year()
  data = data[['organismNameOrig', 'organismCodeOrig']].drop_duplicates()

  # Create the acronym builder
  builder = AcronymBuilder()

  # Get values and acronyms
  values = data['organismNameOrig'].values
  acronyms = data['organismCodeOrig'].values

  print(pd.DataFrame(values, acronyms))

  # Fit builder
  builder.fit(values=values, acronyms=acronyms)

  # Transform
  print(builder.transform(['enterococcus sp.', 'pseudomonas sp.']))

  # Inverse transform
  print(builder.inverse_transform(['ENTC', 'PSEUD']))


  # ---------------------
  # Example II
  # ---------------------
  # This example shows how the existing acronym generation approach 
  # fails for the terms 'feo guapo' and 'feo gordo' which produce the 
  # same acronym.
  # Create data
  a = ''
  b = 'feo'
  c = 'feo   guapo'
  d = 'feo guapo tonto'
  e = 'feo gordo'

  # Create the acronym builder
  builder = AcronymBuilder()

  # Get values and acronyms
  values = [a, b, c, d, e]
  acronyms = [acronym(name) for name in values]

  # Fit builder
  builder.fit(values=values, acronyms=acronyms)



# Libraries






if __name__ == '__main__':

    # Libraries generic
    import time
    import glob
    import pandas as pd

    # Libraries specific
    from pyamr.datasets.clean import clean_basic
    from pyamr.datasets.clean import clean_clwsql008
    from pyamr.datasets.clean import clean_legacy

    # -------------------------------------------
    # Load data
    # -------------------------------------------
    # Define tuples
    tuples = [
        ('./nhs/clwsql008', clean_clwsql008),
        #('./nhs/legacy', clean_legacy),
        #('./nhs/test', clean_clwsql008),
        #('./nhs/test2', clean_legacy)
    ]

    # Combined data
    combined = []

    # For each tuple
    for path, f_clean in tuples:
        # Debug information
        print("Loading... {0}".format(path))
        # Load data (multiple files)
        data = pd.concat([pd.read_csv(f, #nrows=100,
            encoding="ISO-8859-1", engine='c')
                for f in glob.glob(path + "/*.csv")])
        # Clean data
        data = f_clean(data, clean_microorganism=True)
        # Combine
        combined.append(data)

    # Concatenate
    data = pd.concat(combined)

    # It was already within cleaners.
    data = clean_basic(data)

    # Show
    print("\nData:")
    print(data)
    print("\nColumns:")
    print(data.dtypes)


    # -------------------------------------------------
    # Quick and dirty for patient (improve)
    # -------------------------------------------------
    data = data.rename(columns=\
        {'patient_hos_number': 'patient_name'})

    # -------------------
    # Create registries
    # -------------------
    # Information
    print("Creating registries...")

    # Libraries
    from pyamr.datasets.registries import Registry
    from pyamr.datasets.registries import MicroorganismRegistry
    from pyamr.datasets.registries import AntimicrobialRegistry

    # Create registries
    regm = Registry(keyword='method', subset=None).fit(data)
    regx = Registry(keyword='specimen',
        subset=['name', 'code', 'description']).fit(data)
    regs = Registry(keyword='sensitivity').fit(data)
    regp = Registry(keyword='patient').fit(data)
    rego = MicroorganismRegistry(keyword='microorganism').fit(data)
    rega = AntimicrobialRegistry(keyword='antimicrobial').fit(data)

    # Show information
    print("\nSensitivity:")
    print(regs.getr())
    print("\nMethod:")
    print(regm.getr())
    print("\nSpecimen:")
    print(regx.getr())
    print("\nMicroorganism:")
    print(rego.getr())
    print("\nAntimicrobial:")
    print(rega.getr())
    print("\nPatient:")
    print(regp.getr())

    # -------------------------------------------------
    # Format susceptibility test records
    # -------------------------------------------------
    # Information
    print("Formatting susceptibility test records...")

    # Add all the ids
    for r in [regm, regx, regs, rego, rega, regp]:
        data['%s_id' % r.keyword] = \
            r.replace(data['%s_name' % r.keyword],
                key='name', value='id')

    # Why this is soooo slow!
    #for r in [regm, regx, regs, rego, rega, regp]:
    #    data = r.transform(data)


    # -------------------------------------------------
    # Save
    # -------------------------------------------------
    # Information
    print("Creating folder...")

    # Specific
    import csv
    from pathlib import Path

    # Time
    timestr = time.strftime("%Y%m%d-%H%M%S")

    # Define path
    path = Path('./%s' % timestr)

    # Create path if it does not exist
    path.mkdir(parents=True, exist_ok=True)

    # Save registries
    regx.getr().to_csv(path / 'specimen.csv', index=False)
    regs.getr().to_csv(path / 'sensitivity.csv', index=False)
    regm.getr().to_csv(path / 'method.csv', index=False)
    rego.getr().to_csv(path / 'microorganism.csv', index=False)
    rega.getr().to_csv(path / 'antimicrobial.csv', index=False)
    regp.getr().to_csv(path / 'patient.csv', index=False)

    # ----------
    # Filter
    # ----------
    # We have to remove those dates in which the date_received is none
    # because otherwise we cannot group the data by year to store it
    # in different files. In addition, the others are also required to
    # have a meaningful susceptibility test record.
    data = data.dropna(how='any',
                       subset=['date_received',
                               'specimen_code',
                               'microorganism_name',
                               'antimicrobial_name',
                               'sensitivity_name'])

    data = data.drop_duplicates()
    data = data.reset_index(drop=True)
    data['id'] = data.index + 1

    # ----------
    # Save
    # ----------
    # Columns
    keep = ['id',
            'date_received',
            'date_outcome',
            'patient_hos_number',
            'laboratory_number',
            'specimen_code',
            'specimen_name',
            'specimen_description',
            'microorganism_code',
            'microorganism_name',
            'antimicrobial_code',
            'antimicrobial_name',
            'method_code',
            'method_name',
            'sensitivity_code',
            'sensitivity_name',
            'mic',
            'reported',
            'microorganism_id',
            'antimicrobial_id',
            'sensitivity_id',
            'method_id',
            'specimen_id',
            'patient_id']

    # Filter
    data = data[[c for c in keep if c in data]]

    # Create grouper
    grouper = pd.Grouper(key='date_received', freq='Y')

    # .. note: To verify that the date_format do not check
    #          the excels since the format the dates to
    #          by yyy/mm/dd, instead check the raw csvs.

    # Save susceptibility grouped
    for n, g in data.groupby(grouper):
        # Create filename
        filename = "susceptibility-%s.csv" % n.strftime('%Y')
        # Save
        g.to_csv(path / filename, index=False,
                 date_format='%Y-%m-%d %H:%M:%S',
                 quoting=csv.QUOTE_ALL)  # QUOTE_NONNUMERIC

    # Logging
    #logger.info('The results have been saved in: %s' % path)



    import sys
    sys.exit()











    # --------------------------------------------
    # Create registries
    # --------------------------------------------
    # Library
    from pyamr.datasets.registries import create_registry
    
    # Create registries
    reg_patient = create_registry(data, keyword='patient')
    reg_specimen = create_registry(data, keyword='specimen')
    reg_sensitivity = create_registry(data, keyword='sensitivity')
    reg_method = create_registry(data, keyword='method')
    reg_microorganism = create_registry(data, keyword='microorganism')
    reg_antimicrobial = create_registry(data, keyword='antimicrobial')

    # Show
    print("\nMethod:")
    print(reg_method)
    print("\nSensitivity:")
    print(reg_sensitivity)
    print("\nSpecimen:")
    print(reg_specimen)
    print("\nMicroorganism:")
    print(reg_microorganism)
    print("\nAntimicrobial:")
    print(reg_antimicrobial)
    print("\nPatients:")
    print(reg_patient)

    # Fill missing gram stain
    reg_microorganism.gram_stain = reg_microorganism.gram_stain.fillna('u')


    # -------------------------------------------------
    # Complete susceptibility records
    # -------------------------------------------------
    # Helper methods
    def drop_y(df):
        to_drop = [x for x in df if x.endswith('_y')]
        return df.drop(to_drop, axis=1)

    def rename_x(df):
        to_rename = {c: c.rstrip('_x')
            for c in df.columns if c.endswith('_x')}
        return df.rename(columns=to_rename)

    # Merge
    data = data.merge(reg_sensitivity,
        how='left', on='sensitivity_name')
    data = data.merge(reg_method,
        how='left', on='method_code')
    data = data.merge(reg_specimen,
        how='left', on=['specimen_code', 'specimen_description'])
    data = data.merge(reg_patient,
        how='left', on='patient_hos_number')
    data = data.merge(reg_microorganism[['microorganism_name',
                                         'microorganism_id']],
        how='left', on='microorganism_name')
    data = data.merge(reg_antimicrobial[['antimicrobial_name',
                                         'antimicrobial_id']],
        how='left', on='antimicrobial_name')

    # Drop and rename
    data = drop_y(data)
    data = rename_x(data)

    print(data)
    print(data.count())


    # -----------------------------------
    # Save to MySQL
    # ----------------------------------
    # Libraries
    from sqlalchemy import create_engine

    # Constants
    user = 'root'
    pwd = 'toor'
    host = 'localhost'
    schema = 'epicimpoc-test'
    fmt = 'mysql+pymysql://{user}:{pwd}@{host}/{schema}'

    # Create connection
    db_connection = create_engine(fmt.format(user=user,
        pwd=pwd, host=host, schema=schema))

    # Columns to keep
    keep = ['date_created', 'date_updated', 'date_received', 'date_outcome',
            'laboratory_number', 'mic', 'reported', 'antimicrobial_id',
            'method_id', 'microorganism_id', 'patient_id', 'sensitivity_id',
            'specimen_id']

    # Susceptibility tests
    aux = data.copy(deep=True)
    aux['date_created'] = pd.to_datetime('today')
    aux['date_updated'] = pd.to_datetime('today')
    aux = aux[keep]

    # Save
    aux.to_sql(name='microbiology_susceptibilitytest',
               con=db_connection,
               if_exists='replace',
               index=False)

    print(aux)

    import sys
    sys.exit()

    def sql_prepare_lookup(df):
        """Helper method to prepare lookup tables"""
        aux = df.copy(deep=True)
        # Remove prefixes
        aux.columns = [c.split('_', 1)[-1] for c in aux.columns]
        # Add missing columns
        aux['date_created'] = pd.to_datetime('today')
        aux['date_updated'] = pd.to_datetime('today')
        if not 'description' in aux:
            aux['description'] = ''
        if not 'is_visible' in aux:
            aux['is_visible'] = True
        # Fill na with '' so null=False.
        aux = aux.fillna(aux.dtypes.replace({'O': ''}))
        # Keep
        aux = aux[['id', 'name', 'code', 'description',
            'date_created', 'date_updated', 'is_visible']]
        aux = aux.drop_duplicates()
        # Return
        return aux

    LOOP = [#(reg_sensitivity, 'microbiology_sensitivity'),
            #(reg_specimen, 'microbiology_specimen'),
            (reg_method, 'microbiology_method'),
            #(reg_patient, 'microbiology_patient'),
            (reg_antimicrobial, 'microbiology_antimicrobial'),
            (reg_microorganism, 'microbiology_microorganism')]

    for df, name in LOOP:
        try:
            print("Importing.... %s" % name)
            sql_prepare_lookup(df) \
                .to_sql(name=name,
                    con=db_connection,
                    if_exists='append',
                    index=False)
        except Exception as e:
            print(e)



    import sys
    sys.exit()


    # -------------------------------------------------
    #
    # -------------------------------------------------
    # Specific
    from pathlib import Path

    # Time
    timestr = time.strftime("%Y%m%d-%H%M%S")

    # Define path
    path = Path('./%s' % timestr)

    # Create path if it does not exist
    path.mkdir(parents=True, exist_ok=True)

    # Save registries
    reg_patient.to_csv(path / 'patients.csv', index=False)
    reg_specimen.to_csv(path / 'specimens.csv', index=False)
    reg_sensitivity.to_csv(path / 'sensitivities.csv', index=False)
    reg_method.to_csv(path / 'methods.csv', index=False)
    reg_microorganism.to_csv(path / 'microorganisms.csv', index=False)
    reg_antimicrobial.to_csv(path / 'antimicrobial.csv', index=False)

    import sys
    sys.exit()

    # -------------------------------------------------
    # Create susceptibility test record from registries
    # -------------------------------------------------
    data = data.merge(reg_specimen['specimen_code'],
            how='left', left_on='specimen_code',
            right_on='specimen_code')

    data = data.merge(reg_sensitivity,
            how='left', left_on='sensitivity_name',
            right_on='sensitivity_name')

    data = data.merge(reg_method['method_code'],
            how='left', left_on='method_code',
            right_on='method_code')

    keep = ['date_received',
            'date_outcome',
            'laboratory_number',
            'specimen_id',
            'sensitivity_id',
            'method_id']

    # Show
    print(data[keep])


# Libraries
import numpy as np
from matplotlib import colors

class MidpointNormalize(colors.Normalize):
    """Normalise the colorbar so that diverging bars
       work there way either side from a prescribed
       midpoint value)

    Example
    -------
        > MidpointNormalize(midpoint=0., vmin=-100, vmax=100)

    """

    def __init__(self, vmin=None,
                       vmax=None,
                       midpoint=None,
                       clip=False):
        """Constructor
        """
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        """Call

        .. note: Ignoring masked values and all kind of edge cases
                 to keep it simple.
        """
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin,
                self.midpoint,
                self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y), np.isnan(value))


def hlinebgplot(ax, right, yv, bg=None):
    """This function adds a vertical line and background

    Parameters
    ----------

    Returns
    -------
    """
    # Libraries
    import seaborn as sns

    # Plot line
    ax.plot((-1, right), (yv, yv), color='gray',
            linestyle='--', linewidth=1, zorder=0)

    # Plot background.
    if bg is not None:
        cb = sns.color_palette("Set2", 10)[1]
        ax.fill_between([-1, right], [yv, yv], [bg, bg],
                        zorder=0, alpha=0.2, color=cb)

def vlinebgplot(ax, top, xv, bg=None):
    """This function adds a vertical line and background

    Parameters
    ----------
    ax: matplotlib axes
    top: float
        The max y value
    xv: float
        The x value
    bg: boolean
        Whether to include a background

    Returns
    -------
    """
    # Libraries
    import seaborn as sns

    # Plot line.
    ax.plot((xv, xv), (-1, top), color='gray',
            linestyle='--', linewidth=1, zorder=0)

    # Plot background.
    if bg is not None:
        cb = sns.color_palette("Set2", 10)[1]
        ax.fill_between([xv, bg], [-1, -1], [top, top],
                        zorder=0, alpha=0.1, color=cb)


def scalar_colormap(values, cmap, vmin, vmax):
    """This method creates a colormap based on values.

    Parameters
    ----------
    values : array-like
      The values to create the corresponding colors

    cmap : str
      The colormap

    vmin, vmax : float
      The minimum and maximum possible values

    Returns
    -------
    scalar colormap
    """
    # Library
    import matplotlib as mpl
    import seaborn as sns

    # Create scalar mappable
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
    mapper = mpl.cm.ScalarMappable(norm=norm, cmap=cmap)
    # Gete color map
    colormap = sns.color_palette([mapper.to_rgba(i) for i in values])
    # Return
    return colormap


##############################################################################
# Author: Bernard Hernandez
# Filename: adfuller.py
#
# Description : This file contains a wrapper for the adfuller module.
###############################################################################
# https://mkaz.blog/code/python-string-format-cookbook/
# Future
from __future__ import division

# Libraries
import sys
import numpy as np
import pandas as pd

# Import base wrapper
from pyamr.core.stats.wbase import BaseWrapper


class ADFWrapper(BaseWrapper):
    """The ADF wrapper.

    The Augmented Dickey-Fuller test can be used to test for a unit root in
    a univariate process in the presence of serial correlation. It tests the
    null hypothesis that a unit root is present in a time series sample. The
    alternative hypothesis is different depending on which version of the test
    is used, but is usually stationarity or trend-stationarity. The more
    negative the statistic, the stronger the rejection of the hypothesis that
    there is a unit root at some level of confidence.

    Notes
    -----
    H0: The series has a unit root => Non-stationary
    H1: The series has no unit root => Stationary / Trend-Stationary

    If p-value > 0.05: Failed to reject H0.
    If p-value <= 0.05: Reject H0.

    The absence of unit root is not a proof of non-stationarity. As such, it
    is also possible to use the Kwiatkowski–Phillips–Schmidt–Shin (KPSS) test
    to identify the existence of an underlying trend which can also be removed
    to obtain a stationary process. These are called trend-stationary processes.

    In both, unit-root and trend-stationary processes, the mean can be increasing
    or decreasing over time; however, in the presence of a shock, trend-stationary
    processes revert to this mean tendency in the long run (deterministic trend)
    while unit-root processes have a permanent impact (stochastic trend).

    [1] https://machinelearningmastery.com/time-series-data-stationary-python/
    [2] https://www.statsmodels.org/stable/examples/notebooks/generated/stationarity_detrending_adf_kpss.html

    .. todo: .
    """

    def _identifier(self):
        """Description of the model"""
        return "%s(%s)" % (self._name, self.regression)

    # --------------------------------------------------------------------------
    #                             new methods
    # --------------------------------------------------------------------------
    def is_stationary(self, alpha=0.05):
        """This method returns a boolean with the stationarity outocme.

        Parameters
        ----------
        alpha : float
          The statistical significance

        Returns
        -------

        """
        return True if self._result['pvalue'] <= alpha else False

    def stationarity(self, alpha=0.05):
        """This method returns the stationarity outcome.

        Parameters
        ----------
        alpha : float
          The stastistical significance

        Returns
        -------

        """
        return 'stationary' if self.is_stationary(alpha) else 'non-stationary'

    # --------------------------------------------------------------------------
    #                          overriden methods
    # --------------------------------------------------------------------------
    def evaluate(self, **kwargs):
        """Evaluates the model.
        """
        # Create dictionary
        d = {}

        # Basic statistics
        d['statistic'] = self._raw[0]
        d['pvalue'] = self._raw[1]
        d['nlags'] = self._raw[2]
        d['nobs'] = self._raw[3]
        d['stationary'] = self._raw[1] < 0.05

        # Format critical value array.
        for key, value in self._raw[4].items():
            d['criticalvalue_%s' % key] = value

        # Return
        return d

    def as_summary(self, alpha=0.05, verbose=1, **kwargs):
        """This method creates the summary to display.

        .. note: Use self.__dict__ to pass all the parameters
                 to format? Note however that the parameters
                 used are in self._result.
        """
        # Symbols
        alpha_symbol = '\u03B1'

        # Template start
        summary = '    adfuller test stationarity ({r})    \n'
        summary += "=======================================\n"
        summary += "statistic:           {statistic:>18.3f}\n"
        summary += "pvalue:              {pvalue:>18.5f}   \n"
        summary += "nlags:               {nlags:>18.0f}    \n"
        summary += "nobs:                {nobs:>18.0f}     \n"
        summary += "stationarity ({s}={alpha}): {stationarity:>16s}\n"

        if verbose > 5:
            # Add critical values
            summary += "---------------------------------------\n"
            summary += "critical value (1%):  {cc1:>17.5f} \n"
            summary += "critical value (5%):  {cc5:>17.5f} \n"
            summary += "critical value (10%): {cc10:>17.5f}\n"

        if verbose > 7:
            # Add hypothesis description
            summary += "---------------------------------------\n"
            summary += "H0:                    Exists unit-root\n"
            summary += "H1:                 No Exists unit-root\n"
            summary += "pvalue <= {s}:                  Reject H0\n"

        # Template end
        summary += "======================================="

        # Format
        summary = summary.format(r=self.regression,
            statistic=self.statistic, pvalue=self.pvalue,
            nlags=self.nlags, nobs=self.nobs, alpha=alpha,
            stationarity=self.stationarity(alpha),
            cc1=self._result['criticalvalue_1%'],
            cc5=self._result['criticalvalue_5%'],
            cc10=self._result['criticalvalue_10%'],
            s=alpha_symbol)

        # Return
        return summary





if __name__ == '__main__':

    # Libraries
    import pandas as pd

    # Import specific
    from statsmodels.tsa.stattools import adfuller

    # ----------------------------
    # set basic configuration
    # ----------------------------
    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 150)
    pd.set_option('display.precision', 4)

    # ----------------------------
    # create data
    # ----------------------------
    # Constants
    length = 100
    offset = 100
    slope = 10

    # Create time-series.
    x = np.arange(length)
    y = np.random.rand(length) * slope + offset

    # ----------------------------
    # create adfuller object
    # ----------------------------
    # Create object
    adf = ADFWrapper(estimator=adfuller).fit(x=y, regression='ct')

    # Print series.
    print("\n")
    print(adf.as_series())

    # Print summary.
    print("\n")
    print(adf.as_summary())
    print(adf.as_summary(verbose=10))

    # Print identifier.
    print("\n")
    print(adf._identifier())

    # -----------------
    # Save and load
    # -----------------
    # File location
    # fname = '../examples/saved/adfuller-sample.pickle'

    # Save
    # adf.save(fname=fname)

    # Load
    # adf = ADFWrapper().load(fname=fname)

    # -----------
    # Grid search
    # -----------
    # Create wrapper
    adf = ADFWrapper(adfuller)

    # Grid parameters.
    grid_params = {'x': [y], 'regression': ['c', 'n', 'ct']}

    # Get wrappers.
    lwrappers = adf.grid_search(grid_params=grid_params, verbose=1)

    # Get summary.
    summary = adf.from_list_dataframe(lwrappers, flabel=True)

    # Plot result.
    print("\n")
    print(summary.T)



# Libraries
import re
import collections
import numpy as np
import pandas as pd

# -------------------------------------------------------------------
# Constants
# -------------------------------------------------------------------
SENSITIVITY_CODE_REPLACE = {
    'ss': 's',
}


ANTIMICROBIAL_CODE_REPLACE = {
    'AAUGU': 'AAUG'
}

MICROORGANISM_CODE_REPLACE = {
    'ACINE2': 'ACINE',
    'CNS2': 'CNS',
    'CNS3': 'CNS',
    'ECOL2': 'ECOL',
    'KPN2': 'KPNE',
    'PAER2': 'PAER',
    'SAUR2': 'SAUR',
    'LFC2': 'LFC',
    'COLIF2': 'COLIF',
    'COLI2': 'COLIF',
    'ENTAE2': 'EAER',
    'ENTAE': 'EAER',
    'VRE': 'ENTC',
    'MRSA': 'SAUR',
    'MCNS': 'CNS',
    'A_ECOLI': 'ECOL',
    'A_ENTEROBA': 'ENTB',
    'A_SVIRIDANS': 'VIRST',
    'A_CLUSITANI': 'CLUS',
    'A_CDUBLINIE': 'CDUB',
    'A_CPSEUDODI': 'CPSEU'
}

ANTIMICROBIAL_NAME_MAP = {
    '\([^)]*\)': '',  # Remove everything between ()
    '\s{2,}': ' ',  # Remove duplicated spaces
    'gentamicin 200': 'gentamicin'
}

MICROORGANISM_NAME_MAP = {
    # Replace basic
    'strep\.': 'streptococcus ',
    'staph\.': 'staphylococcus ',
    'species': 'sp.',
    'sp.($| )': '',
    'sp(.)?($| )': ' ',
    'second': '',
    'third': '',
    '2nd': '',
    '3rd': '',
    # Specific
    '\*\*\* mrsa \*\*\* isolated': 'staphylococcus aureus',

}

def invert(d):
    return {v:k for k,v in d.items()}

#def invert(d):
#    return reversed(list(d.items()))

# .. note: Keep everything lowercase because all the
#          columns are str.lower() before doing any
#          formatting/replacement in clean_common.
#
# .. note: Using an ordered dict. Thus, when inverting
#          the dictionaries, for repeated entries
#          (e.g. sensitive SS and S) the latter will
#          be used.
SPECIMEN_CODE_MAP = {
    'URNCUL': 'URICUL'
}

SPECIMEN_NAME_MAP = {
    'Urine Micro': 'Urine Culture'
}

# '9MRSN': 'MRSCUL',
# 'URINE CULTURE': 'URICUL',
# 'WOUND CULTURE': 'WOUCUL',
# 'BLOOD CULTURE': 'BLDCUL',
# 'SPUTUM CULTURE': 'SPTCUL',
##'CSF CULTURE': 'CSFCUL',
# 'EYE CULTURE': 'EYECUL',
# 'GENITALCUL': 'GENCUL',
# 'NEONATAL SCREEN': 'NEOCUL',

METHOD_CODE_MAP = {
    'DD': 'Disk Difussion',
    'PHO': 'Public Health Laboratory',
    'MIC': 'Minimum Inhibitory Concentration',
    'MASTU': 'Microscopy-Based Antimicrobial Susceptibility Testing',
}

SENSITIVITY_CODE_MAP = collections.OrderedDict({
    'ss': 'sensitive',
    's': 'sensitive',
    'r': 'resistant',
    'i': 'intermediate',
    'nd': 'not done',
    'hr': 'highly resistant',
    '<<do not report>>': 'hide',
    'hide': '<<do not report>>',
    'validation fix entry': 'fix',
    'fix': 'validation fix entry',
})

# Note that they will be executed in order and thus
# order matters. The changes on the first expression
# will affect the cells that will be used in the next
# iteration
REGEX_MAP = {
    '\([^)]*\)': '',        # Remove everything between ()
    '(\s)?\-(\s)?': '-',    # Remove spaces before after hyphen
    'species': '',          # Rename species for next regexp
    'o157': '',             # Remove scherichia coli o157
    'sp(\.)?(\s|$)+': ' ',  # Remove sp from word.
    'sp..': ' ',            # Remove sp.. <--- HOW TO DO IT WITH PREVIOUS!
    'strep(\.|\s|$)': 'streptococcus ',  # Complete
    'staph(\.|\s|$)': 'staphylococcus ', # Complete
    'staphylococci': 'staphylococcus',   # Correction (add mixed? group?)
    'streptococci': 'streptococcus',     # Correction (add mixed? group?)
    '\s+': ' ',             # Remove duplicated spaces.
}

REGEX_MAP_BASIC = {
    '\s+': ' ',  # Remove duplicated spaces.
}


# -----------------------------------------------------------
# Helper methods
# -----------------------------------------------------------

def hyphen_before(x, w):
    """Ensures hyphen between words is correct.

    Parameters
    ----------
    x: string
        The string to format
    w: string
        The word preceded by hyphen.

    Returns
    -------
    string
        The formatted string
    """
    # Ensure it is a string
    if not isinstance(x, str):
        return x
    # Create expression
    regexp = re.compile(r'(\S*)(\s+)(%s)(\W)+'%w)
    # Return
    return re.sub(regexp, r'\1-\3 ', x)


def word_to_start(x, w, pos='start', verbose=0):
    """Moves the word within the string.

    Parameters
    ----------
    x: string
        The string to format
    w: word
        The word to relocate within the string.
    pos: string, default start
        The position to insert the word. The possible options
        are start (at the beginning) or end (at the end) of the
        string.
    verbose: int
        Level of verbosity

    Returns
    -------
    string
        Formatted string
    """
    # Ensure it is a string
    if not isinstance(x, str):
        return x
    # Create regular expression
    regexp = re.compile(\
        r'(.*|^)(\W|^)%s(\W|$)(.*|$)' % w,
        flags=re.IGNORECASE)
    # Return value if it does not fit.
    if not bool(re.match(regexp, x)):
        return x
    # Return
    if pos == 'start':
        return '%s ' % w + x.replace(w, '')
    if pos == 'end':
        return x.replace(w, '') + ' %s' % w

def string_replace(series, remove={}):
    """This method corrects the strings.

    Parameters
    ----------
    series:

    remove:

    Returns
    -------
    """
    # Format (lower)
    series = series.str.lower()
    # Do str replacements
    for k, v in remove.items():
        series = series.str.replace(k, v)
    # Format (strip)
    series = series.str.strip()
    # Return
    return series



# ----------------------------------------------------
# Main cleaners
# ----------------------------------------------------
def clean_basic(data):
    """Performs the basic cleaning.

    1. Everything to lowercase
    2. Remove spaces begin/end (strip)
    3. Remove duplicate spaces (regexp)
    4. Remove duplicates

    Parameters
    ----------
    data: pd.DataFrame
        The data to clean.

    Returns
    -------
    pd.DataFrame
        The cleaned data
    """
    # Copy dataframe
    data = data.copy(deep=True)

    # Put everything to lowercase
    data = data.applymap(lambda x: x.lower()
        if isinstance(x, str) else x)

    # Drop all spaces
    data = data.applymap(lambda x: x.strip()
        if isinstance(x, str) else x)

    # Drop extra spaces
    #{'\s+': ' '}  # Remove duplicated spaces.

    # Basic formatting
    data = data.drop_duplicates()

    # Return
    return data


def clean_format(data):
    """Final formatting..."""

    aux = data.copy(deep=True)

    # Format title
    for c in ['patient_name',
              'patient_surname']:
        if c in aux:
            aux[c] = aux[c].str.title()

    # Format lower (not needed)
    for c in ['antimicrobial_name',
              'microorganism_name',
              'sensitivity_name',
              'method_name']:
        if c in aux:
            aux[c] = aux[c].str.lower()

    # Format upper
    for c in ['antimicrobial_code',
              'microorganism_code',
              'sensitivity_code',
              'method_code']:
        if c in aux:
            aux[c] = aux[c].str.upper()

    # Strip strings
    aux = aux.apply(lambda x: x.str.strip() \
        if x.dtype == "object" else x)

    # Format date-times
    for c in ['date_received', 'date_outcome']:
        if c in aux:
            aux[c] = pd.to_datetime(aux[c], errors='coerce')

    # Drop duplicates
    aux = aux.drop_duplicates()


def clean_clwsql008(data, clean_microorganism=True):
    """Performs cleaning for clwsql008 data

    1. rename columns
    2. clean basic
    3. correct issue with sensitivities
    4. correct issue with date_received

    Parameters
    ----------
    data: pd.DataFrame
        The data to clean.

    Returns
    -------
    pd.DataFrame
        The cleaned data
    """

    # ---------------------------------
    # Constants
    # ---------------------------------
    # Rename columns
    rename = {
        'DiagnosticTestID': 'uuid',
        'ReceiveDate': 'received_date',
        'ReceiveTime': 'received_time',
        'PtNumber': 'patient_hos_number',
        'AccNumber': 'laboratory_number',
        'BatTstCode': 'specimen_code',
        'OrderName': 'specimen_name',
        'SpecType': 'specimen_description',
        'OrgPieceCounter': 'microorganism_piece_counter',
        'OrgCode': 'microorganism_code',
        'Organism': 'microorganism_name',
        'DrugCode': 'antimicrobial_code',
        'AntiBiotic Name': 'antimicrobial_name',
        'SensMethod': 'method_code',
        'Sensitivity': 'sensitivity_name',
        'MIC': 'mic',
        'Reported': 'reported',
        'FinalDate': 'date_outcome'
    }

    # There are both code and names merged.
    sensitivity_code_map = {
        's': 'sensitive',
        'ss': 'sensitive',
        'r': 'resistant',
        'i': 'intermediate',
        'nd': 'not done',
        'hr': 'highly resistant',
        'hide': 'hide',
        'fix': 'validation fix entry'
    }

    sensitivity_name_map = {
        'sensitive': 's',
        'resistant': 'r',
        'intermediate': 'i',
        'not done': 'nd',
        'highly resistant': 'hr',
        'hide': 'hide',
        '<<do not report>>': 'hide',
        'validation fix entry': 'fix'
    }

    # --------------------------
    # Method
    # --------------------------
    # The method codes are given but the method names are not
    # included. We could use this opportunity to set their
    # values
    data = data.rename(columns=rename)
    data = clean_basic(data)
    #data = data.convert_dtypes()

    # --------------------------
    # Correct sensitivities
    # --------------------------
    # Replace codes with the names
    data['sensitivity_name'] = \
        data.sensitivity_name.replace(sensitivity_code_map)
    # Map names to corresponding codes
    data['sensitivity_code'] = \
        data.sensitivity_name.map(sensitivity_name_map)

    # Add columns (will be replaced later)
    if 'sensitivity_code' in data:
        data['sensitivity'] = data.sensitivity_code

    # --------------------------
    # Add method name
    # --------------------------
    #data['method_name'] = data.method_code


    # --------------------------
    # Clean microorganism
    # --------------------------
    if clean_microorganism:
        # Create registry
        from pyamr.datasets.registries import MicroorganismRegistry

        # Create registry
        rego = MicroorganismRegistry(keyword='microorganism').fit(data)

        # Format microorganism name
        data.microorganism_name = \
            rego.replace(data.microorganism_name,
                key='original', value='name')


    #
    #data['date_received'] = pd.to_datetime(
    #    data['ReceiveDate'] + ' ' +  data['ReceiveTime'], errors='coerce')

    data['date_received'] = pd.to_datetime(
        data['date_received_date'] + ' ' +  data['date_received_time'], errors='coerce')

    data['date_outcome'] = pd.to_datetime(data['FinalDate'], errors='coerce')

    # --------------------------
    # Add date
    # --------------------------
    # Add new column date
    #data['date_received'] = pd.to_datetime(
    #    data.received_date + ' ' + data.received_time, errors='coerce')


    # Format date-times
    for c in ['date_received', 'date_outcome']:
        if c in data:
            data[c] = pd.to_datetime(data[c], errors='coerce')

    # Return data
    return data


def clean_legacy(data, clean_microorganism=True, verbose=10):
    """This method cleans microbiology data from legacy.

    1. Rename columns
    2. clean basic
    3. Add sensitivity code
    4. Correct specimen issue

    Parameters
    ----------
    data: pd.DataFrame
        The data to clean

    Returns
    -------
    pd.DataFrame
        The cleaned data
    """
    # ---------------------------------
    # Constants
    # ---------------------------------
    # Rename columns
    rename = {
        'dateReceived': 'date_received',
        'age': 'age',
        'gender': 'gender',
        'patNumber': 'patient_hos_number',
        'labNumber': 'laboratory_number',
        'orderCode': 'specimen_code',
        'orderName': 'specimen_name',
        'specimenType': 'specimen_description',
        'OrgPieceCounter': 'microorganism_piece_counter',
        'organismCode': 'microorganism_code',
        'organismNameOrig': 'microorganism_name',
        'antibioticCode': 'antimicrobial_code',
        'antibioticName': 'antimicrobial_name',
        'sensitivity': 'sensitivity_name'
    }

    # Map sensitivity names with codes.
    sensitivity_name_map = {
        'sensitive': 's',
        'resistant': 'r',
        'intermediate': 'i',
        'not done': 'nd',
        'highly resistant': 'hr',
        '<<do not report>>': 'hide',
        'validation fix entry': 'fix'
    }


    # --------------------------
    # Method
    # --------------------------
    # The method codes are given but the method names are not
    # included. We could use this opportunity to set their
    # values
    # Drop duplicates
    data = data.rename(columns=rename)
    data = clean_basic(data)
    #data = data.convert_dtypes() # issue with np.nan in replace

    # --------------------------
    # Correct sensitivities
    # --------------------------
    # Create column code
    data['sensitivity_code'] = \
        data.sensitivity_name.map(sensitivity_name_map)

    # --------------------------
    # Correct specimens
    # --------------------------
    # Get those with both name and code
    aux = data[['specimen_code', 'specimen_name']]
    aux = aux.dropna(how='any').drop_duplicates()
    tup1 = zip(aux.specimen_name, aux.specimen_code)
    tup2 = zip(aux.specimen_code, aux.specimen_name)

    # Replace names that appear in code
    data.specimen_code = data.specimen_code.replace(dict(tup1))

    # Fill missing (NaN) names
    data.specimen_name = data.specimen_name \
        .fillna(data.specimen_code.replace(dict(tup2)))

    # --------------------------
    # Add method code/name
    # --------------------------
    data['method_name'] = None
    data['method_code'] = None

    # --------------------------
    # Clean microorganism
    # --------------------------
    if clean_microorganism:
        # Create registry
        from pyamr.datasets.registries import MicroorganismRegistry

        # Create registry
        rego = MicroorganismRegistry(keyword='microorganism').fit(data)

        # Could I do it with fit_transform(data)

        # Format microorganism name
        data.microorganism_name = \
            rego.replace(data.microorganism_name,
                key='original', value='name')

    # Format date-times
    for c in ['date_received', 'date_outcome']:
        if c in data:
            data[c] = pd.to_datetime(data[c], errors='coerce')

    # Return
    return data







































def clean_microorganism(data):
    """This method...."""
    # Copy data
    aux = data.copy(deep=True)

    # Add backup columns
    microorganism_name_original = \
        aux.microorganism_name

    # Put everything to lowercase
    aux = aux.applymap(lambda x: x.lower()
        if isinstance(x, str) else x)

    # Save microorganism name origina
    aux['microorganism_name_original'] = \
        microorganism_name_original

    # Apply regexp mapping
    aux.microorganism_name = \
        aux.microorganism_name.replace(regex=REGEX_MAP)

    # Correct hyphens
    for hp in ['haemolytic']:
        aux.microorganism_name = \
            aux.microorganism_name.str.lower() \
                .transform(hyphen_before, w=hp)

    # Correct order of genus
    for sp in ['enterococcus',
               'staphylococcus',
               'streptococcus',
               'coliform']:
        aux.microorganism_name = \
            aux.microorganism_name.str.lower() \
                .transform(word_to_start, w=sp, pos='start')

    # Correct order of tags
    for sp in ['methicillin resistant',
               'vancomycin resistant',
               'mixed']:
        aux.microorganism_name = \
            aux.microorganism_name.str.lower() \
                .transform(word_to_start, w=sp, pos='end')

    # Apply regexp mapping
    aux = aux.replace(regex=REGEX_MAP)

    # Strip
    aux.microorganism_name = \
        aux.microorganism_name.str.strip()


    keep = ['microorganism_name',
            'microorganism_code',
            'microorganism_name_original']

    # Sort
    aux = aux[keep]
    aux = aux.drop_duplicates(subset=['microorganism_name'])
    aux = aux.sort_values(by='microorganism_name')
    aux = aux.reset_index(drop=True)
    aux.insert(0, 'microorganism_id', aux.index)

    # Return
    return aux



def clean_common(data, verbose=10):
    """This method cleans the microbiology data.

    It assumes the following columns are imputed:
       date_received
       date_outcome
       microorganism_code
       microorganism_name (required = True)
       antimicrobial_code
       antimicrobial_name
       method_code
       method_name
       sensitivity_code
       sensitivity_name

    Parameters
    ----------
    data: pd.DataFrame
        The dataframe to clean

    Returns
    -------
    pd.DataFrame
        The cleaned dataframe
    """
    # ------------------------------
    # Constants
    # ------------------------------
    # Create required columns
    required = [
        'date_received',
        'date_outcome',
        'specimen_name',
        'microorganism_name',
        'antimicrobial_name',
        'method_name',
        'sensitivity_name']

    # Copy data
    aux = data.copy(deep=True)

    # Add required columns
    for c in required:
        if not c in aux.columns:
            aux[c] = None

    # ------------------------------
    # Missing columns and replace
    # ------------------------------
    # Add backup columns
    aux['microorganism_name_original'] = \
        aux.microorganism_name
    aux['antimicrobial_name_original'] = \
        aux.antimicrobial_name

    # Put everything to lowercase
    aux = aux.applymap(lambda x: x.lower()
        if isinstance(x, str) else x)

    for c in ['specimen', 'method', 'sensitivity']:
        c_name = '%s_name' % c
        c_code = '%s_code' % c
        if c_code not in aux:
            aux[c_code] = aux[c_name]
        aux[c_name].fillna(aux[c_code], inplace=True)

    # Verbose
    if verbose > 5:
        print("Formatting... specimen/method/sensitivity.")

    # .. note: It would be also possible to forget about the
    #          codes and create our own acronyms ensuring that
    #          they are unique.
    # Fix codes (ss -> s)
    if 'sensitivity_code' in aux:
        aux.sensitivity_code = \
            aux.sensitivity_code.replace(SENSITIVITY_CODE_REPLACE)
    if 'microorganism_code' in aux:
        aux.microorganism_code = \
            aux.microorganism_code.replace(MICROORGANISM_CODE_REPLACE)
    if 'antimicrobial_code' in aux:
        aux.antimicrobial_code = \
            aux.antimicrobial_code.replace(ANTIMICROBIAL_CODE_REPLACE)

    # Replace
    aux = aux.replace({
        'specimen_name': SPECIMEN_CODE_MAP,
        'specimen_code': invert(SPECIMEN_CODE_MAP),
        'method_name': METHOD_CODE_MAP,
        'method_code': invert(METHOD_CODE_MAP),
        'sensitivity_name': SENSITIVITY_CODE_MAP,
        'sensitivity_code': invert(SENSITIVITY_CODE_MAP)
    })

    # ------------------------------
    # Fixing orgs/abxs names
    # ------------------------------
    # Verbose
    if verbose > 5:
        print("Formatting... microorganism/antimicrobials.")

    # Apply regexp mapping
    aux = aux.replace(regex=REGEX_MAP)

    # Correct hyphens
    for hp in ['haemolytic']:
        aux.microorganism_name = \
            aux.microorganism_name.str.lower() \
                .transform(hyphen_before, w=hp)

    # Correct order of genus
    for sp in ['enterococcus',
               'staphylococcus',
               'streptococcus',
               'coliform']:
        aux.microorganism_name = \
            aux.microorganism_name.str.lower() \
                .transform(word_to_start, w=sp)

    # Apply regexp mapping
    aux = aux.replace(regex=REGEX_MAP)

    # ------------------------------
    # String formatting
    # ------------------------------
    # Verbose
    if verbose > 5:
        print("Formatting... lower/title/upper.")

    # Format title
    for c in ['patient_name',
              'patient_surname']:
        if c in aux:
            aux[c] = aux[c].str.title()

    # Format lower (not needed)
    for c in ['antimicrobial_name',
              'microorganism_name',
              'sensitivity_name',
              'method_name']:
        if c in aux:
            aux[c] = aux[c].str.lower()

    # Format upper
    for c in ['antimicrobial_code',
              'microorganism_code',
              'sensitivity_code',
              'method_code']:
        if c in aux:
            aux[c] = aux[c].str.upper()

    # Strip strings
    aux = aux.apply(lambda x: x.str.strip() \
        if x.dtype == "object" else x)

    # ------------------------------
    # Time formatting
    # ------------------------------
    # Verbose
    if verbose > 5:
        print("Formatting... date_received/date_outcome.")

    # Format date-times
    for c in ['date_received', 'date_outcome']:
        if c in aux:
            aux[c] = pd.to_datetime(aux[c], errors='coerce')

    # We could also use the convert_dtypes, however there is a big
    # issue with the pd.NA values. I think that issue is only if
    # we want to apply replace and there are pd.NA (only works with
    # np.nan) but it should be fine now that all that has been done.
    #aux = aux.convert_dtypes()

    # Remove empty (sensitivity, date_received, ...?)
    #aux = aux.dropna(how='any', subset=[])

    # Drop duplicates
    aux = aux.drop_duplicates()

    # Return
    return aux


def clean_clwsql008_old(data, verbose=10):
    """This method cleans microbiology data from clwsql008.

    .. notes: CLW-SQL-008
      - The sensitivity values found are
        { 'SS', 'R', 'ND', 'I', 'HIDE', 'HR' }

    .. note: Using UTC=True gives an error when using
             django-import-export to import the data in
             the databases (commented).

    Parameters
    ----------
    data: pd.DataFrame
        The dataframe with the data

    Returns
    -------
    pd.DataFrame
        The cleaned dataframe.
    """
    # ---------------------------------
    # Constants
    # ---------------------------------
    # Rename columns
    rename = {
        'DiagnosticTestID': 'uuid',
        'ReceiveDate': 'received_date',
        'ReceiveTime': 'received_time',
        'PtNumber': 'patient_hos_number',
        'AccNumber': 'laboratory_number',
        'BatTstCode': 'specimen_code',
        'OrderName': 'specimen_name',
        'SpecType': 'specimen_description',
        'OrgPieceCounter': 'microorganism_piece_counter',
        'OrgCode': 'microorganism_code',
        'Organism': 'microorganism_name',
        'DrugCode': 'antimicrobial_code',
        'AntiBiotic Name': 'antimicrobial_name',
        'SensMethod': 'method_code',
        'Sensitivity': 'sensitivity_code',
        'MIC': 'mic',
        'Reported': 'reported',
        'FinalDate': 'date_outcome'
    }

    # --------------------------
    # Method
    # --------------------------
    # The method codes are given but the method names are not
    # included. We could use this opportunity to set their
    # values
    # Drop duplicates
    data = data.drop_duplicates()
    data = data.rename(columns=rename)
    #data = data.convert_dtypes()

    # Show
    if verbose > 5:
        print("\n")
        print(data.columns)

    # Add new columns
    data['date_received'] = pd.to_datetime(
        data.received_date + ' ' + data.received_time, errors='coerce')

    # Final formatting
    data = clean_common(data, verbose)

    # Return
    return data



def clean_legacy_old(data, verbose=10):
    """This method cleans microbiology data from legacy.

    .. notes: LEGACY
      - The sensitivities found are ...

    Parameters
    ----------
    data: pd.DataFrame
        The dataframe with the data

    Returns
    -------
    pd.DataFrame
        The cleaned dataframe.
    """
    # ---------------------------------
    # Constants
    # ---------------------------------
    # Rename columns
    rename = {
        'dateReceived': 'date_received',
        'age': 'age',
        'gender': 'gender',
        'patNumber': 'patient_hos_number',
        'labNumber': 'laboratory_number',
        'orderCode': 'specimen_code',
        'orderName': 'specimen_name',
        'specimenType': 'specimen_description',
        'OrgPieceCounter': 'microorganism_piece_counter',
        'organismCode': 'microorganism_code',
        'organismNameOrig': 'microorganism_name',
        'antibioticCode': 'antimicrobial_code',
        'antibioticName': 'antimicrobial_name',
        'sensitivity': 'sensitivity_name'
    }

    # --------------------------
    # Method
    # --------------------------
    # The method codes are given but the method names are not
    # included. We could use this opportunity to set their
    # values
    # Drop duplicates
    data = data.drop_duplicates()
    data = data.rename(columns=rename)
    #data = data.convert_dtypes() # issue with np.nan in replace

    # Show
    if verbose > 5:
        print("\n")
        print(data.columns)

    # Final formatting
    data = clean_common(data, verbose)

    # Return
    return data


def clean_mimic(data):
    """This method...

    .. note: Need to merge datetime for date and datetime for time
             as done in the datablend package if want full info.

    """
    # ---------------------------------
    # Constants
    # ---------------------------------
    # Rename columns
    rename = {
        'subject_id': 'patient_hos_number',
        'micro_specimen_id': 'laboratory_number',
        'spec_type_desc': 'specimen_description',
        'test_seq': 'microorganism_piece_counter',
        'org_name': 'microorganism_name',
        'ab_name': 'antimicrobial_name',
        'test_name' : 'method',
        'interpretation': 'sensitivity_name',
        'chartdate': 'date_received',
        'storedate': 'date_outcome'
    }

    # Replace values
    replace = {
        #'sensitivity': SENSITIVITY_MAP,
        'microorganism_code': MICROORGANISM_CODE_REPLACE,
        'microorganism_name': MICROORGANISM_NAME_MAP,
        'antimicrobial_code': ANTIMICROBIAL_CODE_REPLACE
    }

    # Rename
    data = data.rename(columns=rename)

    # Replace
    data = data.replace(replace)

    # Format
    data.microorganism_name = data.microorganism_name.str.capitalize()
    data.antimicrobial_name = data.antimicrobial_name.str.capitalize()
    data['microorganism_code'] = data.microorganism_name
    data['antimicrobial_code'] = data.antimicrobial_name
    data['specimen_code'] = data.specimen_description

    # Format date
    if 'date_received' in data:

        # Convert to datetime.
        data.date_received = \
            pd.to_datetime(data.date_received, errors='coerce')

        # Ignore those without result
        data = data[data.sensitivity.notna()]

    # Return
    return data


##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division

# Libraries
import sys
import numpy as np
import pandas as pd

# Import base wrapper
from pyamr.core.stats.wbase import BaseWrapper
from pyamr.core.stats.wbase import fargs


class StationarityWrapper(BaseWrapper):
    """This method....


    Types of stationarity in time-series:
      - Trend stationary
      - Seasonal stationary
      - Strictly stationary

    Applies both ADF and KPSS tests.

    ADF            KPSS           Outcome               Note
    ============== ============== ===================== ========================
    Non-Stationary Non-Stationary Non-Stationary
    Stationary     Stationary     Stationary
    Non-Stationary Stationary     Trend-Stationary      Check de-trended series
    Stationary     Non-Stationary Difference-Stationary Check differenced-series
    ============== ============== ===================== ========================

    [1] https://machinelearningmastery.com/time-series-data-stationary-python/
    [2]
    [3]

    """

    # --------------------------------------------------------------------------
    #                          overriden methods
    # --------------------------------------------------------------------------
    def evaluate(self, alpha=0.05, **kwargs):
        """This method initialises the series.
        """
        # Create dictionary.
        d = {}

        # Basic statistics (adfuller).
        d['adf_ct_statistic'] = self._raw['adfuller-ct'][0]
        d['adf_ct_pvalue'] = self._raw['adfuller-ct'][1]
        d['adf_ct_nlags'] = self._raw['adfuller-ct'][2]
        d['adf_ct_nobs'] = self._raw['adfuller-ct'][3]
        for key, value in self._raw['adfuller-ct'][4].items():
            d['adf_ct_criticalvalue_%s' % key] = value

        # Basic statistics (adfuller).
        d['adf_c_statistic'] = self._raw['adfuller-c'][0]
        d['adf_c_pvalue'] = self._raw['adfuller-c'][1]
        d['adf_c_nlags'] = self._raw['adfuller-c'][2]
        d['adf_c_nobs'] = self._raw['adfuller-c'][3]
        for key, value in self._raw['adfuller-c'][4].items():
            d['adf_c_criticalvalue_%s' % key] = value

        # Basic statistics (kpss).
        d['kpss_ct_statistic'] = self._raw['kpss-ct'][0]
        d['kpss_ct_pvalue'] = self._raw['kpss-ct'][1]
        d['kpss_ct_nlags'] = self._raw['kpss-ct'][2]
        for key, value in self._raw['kpss-ct'][3].items():
            d['kpss_ct_criticalvalue_%s' % key] = value

        # Basic statistics (kpss).
        d['kpss_c_statistic'] = self._raw['kpss-c'][0]
        d['kpss_c_pvalue'] = self._raw['kpss-c'][1]
        d['kpss_c_nlags'] = self._raw['kpss-c'][2]
        for key, value in self._raw['kpss-c'][3].items():
            d['kpss_c_criticalvalue_%s' % key] = value

        # Extra parameters.
        d['root_ct_stationary'] = d['adf_ct_pvalue'] <= alpha
        d['root_c_stationary'] = d['adf_c_pvalue'] <= alpha
        d['trend_ct_stationary'] = d['kpss_ct_pvalue'] > alpha
        d['trend_c_stationary'] = d['kpss_c_pvalue'] > alpha

        # Unit root (Range unit root test)
        d['rur_statistic'] = self._raw['rur'][0]
        d['rur_pvalue'] = self._raw['rur'][1]
        for key, value in self._raw['rur'][2].items():
            d['rur_criticalvalue_%s' % key] = value

        # Return
        return d

    def as_summary(self, alpha=0.05):
        """This method creates the summary to display.
        """
        # Create summary.
        summary = '      stationarity (alpha=0.05)   \n'
        summary += '==================================\n'
        summary += '          root           trend    \n'
        summary += '----------------------------------\n'
        summary += 'c   {0!s:>5s} ({1:.3f})   {2!s:>5s} ({3:.3f})\n'
        summary += 'ct  {4!s:>5s} ({5:.3f})   {6!s:>5s} ({7:.3f})\n'
        summary += '=================================='

        # Format
        summary = summary.format(self.root_c_stationary,
                                 self.adf_c_pvalue,
                                 self.trend_c_stationary,
                                 self.kpss_c_pvalue,
                                 self.root_ct_stationary,
                                 self.adf_ct_pvalue,
                                 self.trend_ct_stationary,
                                 self.kpss_ct_pvalue)

        # Return
        return summary

    def fit(self, x, adf_kwargs={}, kpss_kwargs={}, **kwargs):
        """This method studies the stationarity of a given time-series.

        The parameters which can be passed to the adfuller and kpss methods
        are listed below:

        - adfuller_kwargs = {x, maxlag, regression, autolag, store, regresults}
        - kpss_kwargs = {x, regression, lags, store}

        @see statsmodels.tsa.stattools.adfuller
        @see statsmodels.tsa.stattoosl.kpss

        Parameters
        ----------
        x : array-like
          The time series

        adf_kwargs : dict-like
          The parameters to pass to the adfuller function

        kpss_kwargs : dict-like
          The parameters to apss to the kpss function


        Returns
        -------
        object : An StationarityWrapper objects.
        """
        # Library.
        from statsmodels.tsa.stattools import adfuller
        from statsmodels.tsa.stattools import kpss
        from statsmodels.tsa.stattools import range_unit_root_test

        # Empty the class
        self._empty()

        # In this fit a number of scenarios are going to be tested. The
        # term that varies within scenarios is regression, as such, if
        # it is passed it will be deleted.
        adf_kwargs.pop('regression', None)
        kpss_kwargs.pop('regression', None)

        # Update the configuration
        self._config.update({'adf_%s' % k: v for k, v in adf_kwargs.items()})
        self._config.update({'kpss_%s' % k: v for k, v in kpss_kwargs.items()})

        # Initialize raw data.
        self._raw = {'x': x}

        # Compute adfuller and kpss
        self._raw['adfuller-ct'] = adfuller(x=x, regression='ct', **adf_kwargs)
        self._raw['adfuller-c'] = adfuller(x=x, regression='c', **adf_kwargs)
        self._raw['kpss-ct'] = kpss(x=x, regression='ct', **kpss_kwargs)
        self._raw['kpss-c'] = kpss(x=x, regression='c', **kpss_kwargs)
        self._raw['rur'] = range_unit_root_test(x=x)

        print(self._raw['rur'])

        # Evaluate the model
        if self.evaluate:
            self._result = self.evaluate()

        # Save results.
        return self


if __name__ == '__main__':

    # Libraries
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    # ----------------------------
    # set basic configuration
    # ----------------------------
    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 150)
    pd.set_option('display.precision', 4)

    # Set default parameters.
    mpl.rc('lines', linewidth=0.35)
    mpl.rc('xtick', labelsize=6)
    mpl.rc('ytick', labelsize=6)
    mpl.rc('legend', fontsize=6)
    mpl.rc('grid')
    mpl.rc('figure')
    mpl.rc('axes')
    mpl.rc('font', size=7)

    # Font type.
    font = {
        'family': 'monospace',
        'weight': 'normal',
        'size': 6,
    }

    # ----------------------------
    # create data
    # ----------------------------
    # Constants
    length = 100
    offset = 100
    slope = 4

    # Create variables.
    x = np.arange(length)
    n = np.random.rand(length)

    # Create timeseries.
    y_n = n
    y_c = np.ones(length) * offset
    y_t = x * slope + n
    y_ct = x * slope + offset + n * 20
    y_r = np.concatenate((y_ct[:50], y_ct[50:] - offset))

    # ----------------------------
    # create stationarity objects
    # ----------------------------
    # .. note:: Including the constant series with offset produces
    #           the following error: ValueError: cannot convert float
    #           NaN to integer.

    stationarity_n = StationarityWrapper().fit(x=y_n)
    #stationarity_c = StationarityWrapper().fit(x=y_c)
    stationarity_t = StationarityWrapper().fit(x=y_t)
    stationarity_r = StationarityWrapper().fit(x=y_r)
    stationarity_ct = StationarityWrapper().fit(x=y_ct,
         adf_kwargs={'maxlag': 12, 'autolag': 'BIC'})

    # Print series.
    print("\n")
    print(stationarity_ct.as_series())

    # Print summary.
    print("\n")
    print(stationarity_ct.as_summary())

    # Print identifier.
    print("\n")
    print(stationarity_ct._identifier())

    # ----------------
    # plot
    # ----------------
    # Create figure
    fig, axes = plt.subplots(3, 2, figsize=(10, 4))
    axes = axes.flatten()

    # Plot truth values.
    axes[0].plot(y_n, color='#A6CEE3', alpha=0.5, marker='o',
                 markeredgecolor='k', markeredgewidth=0.5,
                 markersize=2, linewidth=0.75,
                 label=stationarity_n.as_summary())

    #axes[1].plot(y_c, color='#A6CEE3', alpha=0.5, marker='o',
    #             markeredgecolor='k', markeredgewidth=0.5,
    #             markersize=2, linewidth=0.75,
    #             label=stationarity_c.as_summary())

    # Plot truth values.
    axes[2].plot(y_t, color='#A6CEE3', alpha=0.5, marker='o',
                 markeredgecolor='k', markeredgewidth=0.5,
                 markersize=2, linewidth=0.75,
                 label=stationarity_t.as_summary())

    # Plot truth values.
    axes[3].plot(y_ct, color='#A6CEE3', alpha=0.5, marker='o',
                 markeredgecolor='k', markeredgewidth=0.5,
                 markersize=2, linewidth=0.75,
                 label=stationarity_ct.as_summary())

    # Plot truth values.
    axes[4].plot(y_r, color='#A6CEE3', alpha=0.5, marker='o',
                 markeredgecolor='k', markeredgewidth=0.5,
                 markersize=2, linewidth=0.75,
                 label=stationarity_r.as_summary())

    # Add grid
    for ax in axes:
        ax.grid(color='gray', linestyle='--', linewidth=0.2, alpha=0.5)

    # Add legend
    for ax in axes:
        ax.legend(prop=font, loc=2)

    # Study of Stationarity
    plt.suptitle("Study of Stationarity")

    # -----------------
    # Save and load
    # -----------------
    # File location
    # fname = '../examples/saved/stationarity-sample.pickle'

    # Save
    # stationarity_ct.save(fname=fname)

    # Load
    # stationarity_ct = StationarityWrapper().load(fname=fname)

    # Show
    #plt.show()





# -------------------------------------------
# Methods
# -------------------------------------------
def create_registry(data, keyword=None, keep=None):
    """Creates registry from data.

    Parameters
    ----------
    data: pd.DataFrame
        The data

    keyword: string
        The keyword for the columns. All columns starting
        with such keyword will be kept and used for the
        registry.

    keep: list
        The list of columns to keep for the registry
    """
    # Columns to keep
    if keep is None:
        keep = [c for c in data.columns
            if c.startswith(keyword)]

    # Copy data
    reg = data[keep].copy(deep=True)
    reg = reg.drop_duplicates()
    reg = reg.reset_index(drop=True)

    # Add id
    if keyword is not None:
        reg['%s_id' % keyword] = reg.index.values

    # Return
    return reg




if __name__ == '__main__':

    # Libraries generic
    import time
    import pandas as pd

    # Libraries specific
    from pyamr.datasets.load import make_susceptibility
    from pyamr.datasets.load import load_data_nhs

    # -------------------------------------------
    # Load data
    # -------------------------------------------
    # Load data
    data = make_susceptibility()
    data, abxs, orgs = \
        load_data_nhs(folder='susceptibility-v0.0.5')

    # Show
    print("\nData:")
    print(data)
    print("\nColumns:")
    print(data.columns)

    # --------------------------------------------
    # Create registries
    # --------------------------------------------
    # Create registries
    reg_patient = create_registry(data, keyword='patient')
    reg_specimen = create_registry(data, keyword='specimen')
    reg_sensitivity = create_registry(data, keyword='sensitivity')
    reg_method = create_registry(data, keyword='method')
    reg_microorganism = create_registry(data, keyword='microorganism')
    reg_antimicrobial = create_registry(data, keyword='antimicrobial')

    # Show
    print("\nMethod:")
    print(reg_method)
    print("\nSensitivity:")
    print(reg_sensitivity)
    print("\nSpecimen:")
    print(reg_specimen)
    print("\nMicroorganism:")
    print(reg_microorganism)
    print("\nAntimicrobial:")
    print(reg_antimicrobial)
    print("\nPatients:")
    print(reg_patient)

    # -----------------------------------------------
    # Include microorganism/antimicrobial information
    # -----------------------------------------------
    # Libraries
    from pyamr.datasets.registries import MicroorganismRegistry
    from pyamr.datasets.registries import AntimicrobialRegistry

    # Load registry
    mreg = MicroorganismRegistry()
    areg = AntimicrobialRegistry()

    # Create genus and species
    reg_microorganism[['genus', 'species']] = \
        reg_microorganism.microorganism_name \
            .str.capitalize() \
            .str.split(expand=True, n=1)

    # Combine with registry information
    reg_microorganism = mreg.combine(reg_microorganism)
    reg_antimicrobial = areg.combine(reg_antimicrobial)

    # This fix with others in registries
    reg_antimicrobial.antimicrobial_name = \
        reg_antimicrobial.antimicrobial_name.str.lower()

    # Fill missing gram stain
    reg_microorganism.gram_stain = reg_microorganism.gram_stain.fillna('u')

    # -------------------------------------------------
    # Complete susceptibility records
    # -------------------------------------------------
    # Helper methods
    def drop_y(df):
        to_drop = [x for x in df if x.endswith('_y')]
        return df.drop(to_drop, axis=1)

    def rename_x(df):
        to_rename = {c: c.rstrip('_x')
            for c in df.columns if c.endswith('_x')}
        return df.rename(columns=to_rename)

    # Merge
    data = data.merge(reg_sensitivity,
        how='left', on='sensitivity_name')
    data = data.merge(reg_method,
        how='left', on='method_code')
    data = data.merge(reg_specimen,
        how='left', on=['specimen_code', 'specimen_description'])
    data = data.merge(reg_patient,
        how='left', on='patient_hos_number')
    data = data.merge(reg_microorganism[['microorganism_name',
                                         'microorganism_id']],
        how='left', on='microorganism_name')
    data = data.merge(reg_antimicrobial[['antimicrobial_name',
                                         'antimicrobial_id']],
        how='left', on='antimicrobial_name')

    # Drop and rename
    data = drop_y(data)
    data = rename_x(data)

    print(data)
    print(data.count())


    # -----------------------------------
    # Save to MySQL
    # ----------------------------------
    # Libraries
    from sqlalchemy import create_engine

    # Constants
    user = 'root'
    pwd = 'toor'
    host = 'localhost'
    schema = 'epicimpoc-test'
    fmt = 'mysql+pymysql://{user}:{pwd}@{host}/{schema}'

    # Create connection
    db_connection = create_engine(fmt.format(user=user,
        pwd=pwd, host=host, schema=schema))

    # Columns to keep
    keep = ['date_created', 'date_updated', 'date_received', 'date_outcome',
            'laboratory_number', 'mic', 'reported', 'antimicrobial_id',
            'method_id', 'microorganism_id', 'patient_id', 'sensitivity_id',
            'specimen_id']

    # Susceptibility tests
    aux = data.copy(deep=True)
    aux['date_created'] = pd.to_datetime('today')
    aux['date_updated'] = pd.to_datetime('today')
    aux = aux[keep]

    # Save
    aux.to_sql(name='microbiology_susceptibilitytest',
               con=db_connection,
               if_exists='replace',
               index=False)

    print(aux)

    import sys
    sys.exit()

    def sql_prepare_lookup(df):
        """Helper method to prepare lookup tables"""
        aux = df.copy(deep=True)
        # Remove prefixes
        aux.columns = [c.split('_', 1)[-1] for c in aux.columns]
        # Add missing columns
        aux['date_created'] = pd.to_datetime('today')
        aux['date_updated'] = pd.to_datetime('today')
        if not 'description' in aux:
            aux['description'] = ''
        if not 'is_visible' in aux:
            aux['is_visible'] = True
        # Fill na with '' so null=False.
        aux = aux.fillna(aux.dtypes.replace({'O': ''}))
        # Keep
        aux = aux[['id', 'name', 'code', 'description',
            'date_created', 'date_updated', 'is_visible']]
        aux = aux.drop_duplicates()
        # Return
        return aux

    LOOP = [#(reg_sensitivity, 'microbiology_sensitivity'),
            #(reg_specimen, 'microbiology_specimen'),
            (reg_method, 'microbiology_method'),
            #(reg_patient, 'microbiology_patient'),
            (reg_antimicrobial, 'microbiology_antimicrobial'),
            (reg_microorganism, 'microbiology_microorganism')]

    for df, name in LOOP:
        try:
            print("Importing.... %s" % name)
            sql_prepare_lookup(df) \
                .to_sql(name=name,
                    con=db_connection,
                    if_exists='append',
                    index=False)
        except Exception as e:
            print(e)



    import sys
    sys.exit()


    # -------------------------------------------------
    #
    # -------------------------------------------------
    # Specific
    from pathlib import Path

    # Time
    timestr = time.strftime("%Y%m%d-%H%M%S")

    # Define path
    path = Path('./%s' % timestr)

    # Create path if it does not exist
    path.mkdir(parents=True, exist_ok=True)

    # Save registries
    reg_patient.to_csv(path / 'patients.csv', index=False)
    reg_specimen.to_csv(path / 'specimens.csv', index=False)
    reg_sensitivity.to_csv(path / 'sensitivities.csv', index=False)
    reg_method.to_csv(path / 'methods.csv', index=False)
    reg_microorganism.to_csv(path / 'microorganisms.csv', index=False)
    reg_antimicrobial.to_csv(path / 'antimicrobial.csv', index=False)

    import sys
    sys.exit()

    # -------------------------------------------------
    # Create susceptibility test record from registries
    # -------------------------------------------------
    data = data.merge(reg_specimen['specimen_code'],
            how='left', left_on='specimen_code',
            right_on='specimen_code')

    data = data.merge(reg_sensitivity,
            how='left', left_on='sensitivity_name',
            right_on='sensitivity_name')

    data = data.merge(reg_method['method_code'],
            how='left', left_on='method_code',
            right_on='method_code')

    keep = ['date_received',
            'date_outcome',
            'laboratory_number',
            'specimen_id',
            'sensitivity_id',
            'method_id']

    # Show
    print(data[keep])


# Import libraries
import warnings
import pandas as pd


# -------------------------------------------------------------------------
#                            helper methods
# -------------------------------------------------------------------------
def _check_dataframe(dataframe):
    """Ensure the three columns are available.

    It also fills the empty values with zeros. This is valid
    for SARI because the columns just represent the number of
    records in that category. Thus, nan is equivalent to 0.
    """
    # Check there is at least one column.
    if not 'resistant' in dataframe and \
       not 'sensitive' in dataframe and \
       not 'intermediate' in dataframe:
       raise Exception("""
        To compute the SARI at least one of the following
        columns must be present in the DataFrame (resistant,
        sensitive, intermediate)""")

    # Copy
    aux = dataframe.copy(deep=True)

    # Add missing columns
    if not 'resistant' in dataframe:
        aux['resistant'] = 0
    if not 'intermediate' in  dataframe:
        aux['intermediate'] = 0
    if not 'sensitive' in dataframe:
        aux['sensitive'] = 0

    # Fill missing.
    aux.resistant.fillna(0, inplace=True)
    aux.intermediate.fillna(0, inplace=True)
    aux.sensitive.fillna(0, inplace=True)

    # return
    return aux


def sari(dataframe=None, strategy='hard', **kwargs):
    """Computes the sari index.

    Parameters
    ----------
    dataframe: pd.DataFrame
        A dataframe with the susceptibility test interpretations
        as columns. The default strategies used (see below) expect
        the following columns ['sensitive', 'intermediate', 'resistant']
        and if they do not appear they weill be set to zeros.

    strategy: string or func, default='hard'
        The method used to compute sari. The possible options
        are 'soft', 'medium' and 'hard'. In addition, a function
        with the following signature func(dataframe, **kwargs)
        can be passed.

            (i) ``basic`` as R / R+S
            (ii) ``soft``    as R / R+I+S
            (iii) ``medium``  as R+0.5I / R+0.5I+S
            (iv) ``hard``  as R+I / R+I+S

    **kwargs: arguments to pass the strategy function.

    Returns
    -------
    pd.Series
        The resistance index
    """
    # Ensure that exists
    aux = _check_dataframe(dataframe)

    # Extract vectors
    r = aux['resistant']
    i = aux['intermediate']
    s = aux['sensitive']

    # Call the function
    if callable(strategy):
        return strategy(aux, **kwargs)

    # Check strategy.
    if strategy not in ['soft', 'medium', 'hard', 'basic']:
        raise ValueError("""
              The strategy '{0}' is not supported. Please
              use one of the following: soft, medium, hard
              or basic""".format(strategy))

    # Compute
    if strategy == 'hard':
        return (r + i) / (r + i + s)
    elif strategy == 'medium':
        return (r + 0.5*i) / (r + i + s)
    elif strategy == 'soft':
        return r / (r + i + s)
    elif strategy == 'basic':
        return r / (r + s)


class SARI:
    """Single Antimicrobial Resistance Trend
    """
    # Attributes
    c_spe = 'SPECIMEN'
    c_org = 'MICROORGANISM'
    c_abx = 'ANTIMICROBIAL'
    c_dat = 'DATE'
    c_out = 'SENSITIVITY'

    def __init__(self, groupby=[c_spe,
                                c_org,
                                c_abx,
                                c_out]):
        """Constructor.

        Parameters
        ----------
        groupby: list
            The labels of the columns to groupby.

        Returns
        --------
        SARI instance
        """
        self.groupby = groupby


    def rolling(self, dataframe, period, cdate, shift=None):
        """"""
        if shift is None:
            warnings.warn("""
                The input parameter <shift> is None. Thus, the value 
                of the input parameter <period> (%s) has been used."""
                % period)
            shift = period

        # Grouper
        grouper = [pd.Grouper(freq=shift, key=cdate)]
        grouper = grouper + self.groupby

        # Compute frequencies
        freqs = dataframe.groupby(grouper) \
            .size().unstack().reset_index() \
            .set_index(cdate).groupby(grouper[1:-1]) \
            .rolling(window=period, min_periods=1) \
            .sum().fillna(0)

        # Return
        return freqs


    def grouping(self, dataframe, period, cdate):
        """Groups the data.
        """
        # Create grouper
        if hasattr(dataframe[cdate].dt, str(period)):
            grouper = [getattr(dataframe[cdate].dt, period)]
        else:
            grouper = [pd.Grouper(freq=period, key=cdate)]
        grouper = self.groupby + grouper

        # Compute sensitivity counts
        freqs = dataframe.groupby(grouper) \
            .size().unstack(level=-2) \
            .fillna(0)

        # Return
        return freqs


    def compute(self, dataframe, period=None, shift=None, cdate=None,
                return_frequencies=True, **kwargs):
        """Computes single antibiotic resistance index.

        .. todo: Add parameters to rolling!
        .. todo: Place value at the left, center, right of window?
        .. todo: Ensure that works when time gaps present!
        .. todo: Compare period > shift
        .. todo: Warning if dates NaN
        .. todo: Warning if elements in groupby any all NaN!
        .. todo: Warning if not all samples tested with same antimicrobials

        Examples
        --------

        ======= ====== =======================================================
        shift   period description
        ======= ====== =======================================================
        None    None   Uses all data (agnostic to time)
        None    year   Value every year using whole year's data.
        None    2D     Value every 2 days using 2 days data.
        2D      2D     Value every 2 days using 2 days data.
        None    2      ``Invalid``
        2D      2      ``Invalid`` Value every 2 days using 2x2D=4 days data.
        2D      None   ``Invalid`` - period cannot be None (in this case)
        2D      year   ``Invalid`` - period cannot be a year (in this case)
        year    --     ``Invalid`` - shift cannot be a named time.
        2       --     ``Invalid`` - shift cannot be a number.
        ======= ====== =======================================================

        Note that shift=2D and period=2D is equivalent to shift=2D and period=1.

        Parameters
        ----------
        dataframe: pd.DataFrame
            A dataframe with the susceptibility test interpretations
            as columns. The default strategies used (see below) expect
            the following columns ['sensitive', 'intermediate', 'resistant']
            and if they do not appear they weill be set to zeros.

        shift: str
            Frequency (datetime) value to group by when applying a rolling window.

        period: str, int
            If used alone (shift=None) is the value used to create groups (e.g. year).
            The whole data within the groups will be used to compute the metrics. On
            the contrary, when used in combinations with shift, it indicates the
            interval used to compute the metrics (e.g. 2D, 2 times the shift value)

        cdate: string, default=None
            The column that will be used as date.

        return_frequencies: boolean, default=True
            Whether to return the frequencies or just the resistance index.

        strategy: string or func, default='hard'
            The method used to compute sari. The possible options
            are 'soft', 'medium' and 'hard'. In addition, a function
            with the following signature func(DataFrame, **kwargs)
            can be passed.

                (i) ``basic`` as R / R+S
                (ii) ``soft``    as R / R+I+S
                (iii) ``medium``  as R+0.5I / R+0.5I+S
                (iv) ``hard``  as R+I / R+I+S

        **kwargs: arguments to pass the strategy function.

        Returns
        -------
        pd.Series or pd.DataFrame
            The resistance index (pd.Series) or a pd.DataFrame with the
            resistance index (sari) and the frequencies.
        """
        # Copy DataFrame
        aux = dataframe.copy(deep=True)

        # Not allowing period to be a number. The main reason is that the
        # most common interpretation is that scenarios with shift=1D
        # period=2D and shift=1D period=2 should be the same. However, the
        # results are actually different. Because period=2 in rolling will
        # use two adjacent rows without considering time. This introduces
        # inconsistencies where there are time gaps without data.
        if period is not None:
            if not isinstance(period, str):
                raise ValueError("""
                    The input parameter <period> cannot be of %s. Ensure 
                    it is either None or a valid string such as 2D or year.
                    """ % type(period))

        # ------------------------------------------
        # Frequencies
        # ------------------------------------------
        # Compute frequencies
        if period is None and shift is None:
            freqs = aux.groupby(self.groupby) \
                .size().unstack().fillna(0)

        else:

            # Format as datetime
            aux[cdate] = pd.to_datetime(aux[cdate])

            if shift is None:
                freqs = self.grouping(dataframe=aux,
                                      period=period,
                                      cdate=cdate)
            else:
                freqs = self.rolling(dataframe=aux,
                                     period=period,
                                     shift=shift,
                                     cdate=cdate)

        # Remove index name.
        freqs.columns.name = None

        # -------------------
        # Sari
        # -------------------
        # Compute sari
        s = sari(freqs, **kwargs).rename('sari')

        # Return
        if return_frequencies:
            freqs['freq'] = freqs.sum(axis=1)
            freqs['sari'] = sari(freqs, **kwargs)
            return freqs
        return s











if __name__ == '__main__': # pragma: no cover

    # Libraries
    import pandas as pd
    import matplotlib as mpl

    # Specific
    from pyamr.core.sari import SARI

    # Set matplotlib
    mpl.rcParams['xtick.labelsize'] = 9
    mpl.rcParams['ytick.labelsize'] = 9
    mpl.rcParams['axes.titlesize'] = 11
    mpl.rcParams['legend.fontsize'] = 9

    # ----------------------------------
    # Create data
    # ----------------------------------
    # Define susceptibility test records
    data = [
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-04', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],

        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],
        ['2021-01-03', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-04', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],

        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-01', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-02', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-02', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-08', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-09', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],

        ['2021-01-12', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-12', 'URICUL', 'SAUR', 'ACIP', 'intermediate'],
        ['2021-01-13', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-13', 'URICUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-14', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-14', 'URICUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-15', 'URICUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-15', 'URICUL', 'SAUR', 'ACIP', 'sensitive'],
        ['2021-01-16', 'URICUL', 'SAUR', 'ACIP', 'intermediate'],
        ['2021-01-16', 'URICUL', 'SAUR', 'ACIP', 'intermediate'],
    ]

    data = pd.DataFrame(data,
                        columns=['DATE',
                                 'SPECIMEN',
                                 'MICROORGANISM',
                                 'ANTIMICROBIAL',
                                 'SENSITIVITY'])

    # Create SARI instance
    sari = SARI(groupby=['SPECIMEN',
                         'MICROORGANISM',
                         'ANTIMICROBIAL',
                         'SENSITIVITY'])

    # Compute SARI overall
    sari_overall = sari.compute(data,
         return_frequencies=True)

    # Compute SARI temporal (ITI)
    sari_iti = sari.compute(data, shift='2D',
         period='2D', cdate='DATE')

    # Compute SARI temporal (OTI)
    sari_oti = sari.compute(data, shift='1D',
         period='2D', cdate='DATE')

    # Show
    print("\nSARI (overall):")
    print(sari_overall)
    print("\nSARI (iti):")
    print(sari_iti)
    print("\nSARI (oti):")
    print(sari_oti)
    print("\n\n")




    # -----------------------
    # Full test
    # -----------------------
    from itertools import product

    # Define possible values
    values = [2, '2D', None, 'year']
    combos = list(product(values, values))

    # Show
    print("\n\nCombinations of params <shift> and <period>:")

    # Loop
    for i, (shift, period) in enumerate(combos):
        print("%2s/%2s. Computing... shift=%-5s | period=%-5s ==> " % \
              (i+1, len(combos), shift, period), end="")
        try:
            s = sari.compute(data,
                shift=shift, period=period, cdate='DATE')
            print("Ok!")
            #print(s)
            #print("\n\n" + "="*80)
        except Exception as e:
            print(e)






















"""
# Import libraries
import sys
import matplotlib as mpl
import matplotlib.pyplot as plt

# Import specific libraries
from pyamr.datasets import load

# Set matplotlib
mpl.rcParams['xtick.labelsize'] = 9
mpl.rcParams['ytick.labelsize'] = 9
mpl.rcParams['axes.titlesize'] = 11
mpl.rcParams['legend.fontsize'] = 9

# -----------------------
# Load data
# -----------------------
# Load sample data
data = load.dataset_epicimpoc_susceptibility_year(nrows=1000000)

# Keep only relevant columns
data = data[['antibioticCode',
           'organismCode',
           'dateReceived',
           'sensitivity']]


# Filter for two examples
is_org = data['organismCode']=='ECOL'
is_abx = data['antibioticCode'].isin(['AAUG'])
data = data[is_abx & is_org]

# -------------------------
# Create frequency instance
# -------------------------
# Create instance
freq = Frequency(column_antibiotic='antibioticCode',
               column_organism='organismCode',
               column_date='dateReceived',
               column_outcome='sensitivity')


# Compute frequencies daily
daily = freq.compute(data, strategy='ITI',
                           by_category='pairs',
                           fs='1D')

# Compute frequencies monthly
monthly = freq.compute(data, strategy='ITI',
                           by_category='pairs',
                           fs='1M')

# Compute frequencies overlapping
oti_1 = freq.compute(data, strategy='OTI',
                         by_category='pairs',
                         wshift='1D',
                         wsize=90)

# -------------------------
# Create sari instance
# -------------------------
# Create instance
sari_daily = SARI(strategy='hard').compute(daily)
sari_monthly = SARI(strategy='hard').compute(monthly)
sari_oti_1 = SARI(strategy='hard').compute(oti_1)

# -------
# Plot
# -------
# Show comparison for each pair
f, axes = plt.subplots(4, 1, figsize=(15,8))

# Flatten axes
axes = axes.flatten()

# Plot ITI (monthly)
for i,(pair, group) in enumerate(sari_daily.groupby(level=[0,1])):
    group.index = group.index.droplevel([0,1])
    group['sari'].plot(marker='o', ms=3, label=pair,
      linewidth=0.5, markeredgecolor='k', markeredgewidth=0.3,
      ax=axes[0])

# Plot ITI (monthly)
for i,(pair, group) in enumerate(sari_monthly.groupby(level=[0,1])):
    group.index = group.index.droplevel([0,1])
    group['sari'].plot(marker='o', ms=3, label=pair,
      linewidth=0.5, markeredgecolor='k', markeredgewidth=0.3,
      ax=axes[1])

# Plot OTI (daily with size 30)
for i,(pair, group) in enumerate(sari_oti_1.groupby(level=[0,1])):
    group.index = group.index.droplevel([0,1])
    group['sari'].plot(marker='o', ms=3, label=pair,
      linewidth=0.5, markeredgecolor='k', markeredgewidth=0.3,
      ax=axes[2])

# Set legend
for ax in axes:
    ax.legend()
    ax.set_xlabel('')
    ax.grid(True)

# Set titles
axes[0].set_ylabel('Daily')
axes[1].set_ylabel('Monthly')
axes[2].set_ylabel('OTI(1D,90)')

# Show
plt.show()
"""



'''
Author: Bernard Hernandez
Filename:
Date: 
Description: 
'''

# Libraries.
from os import listdir

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import datetime
import glob
import json
import os as os

#-------------------------------------------------------------------------
#                           HELPER METHODS
#-------------------------------------------------------------------------
def files_with_extension(folder, ext):
	""" This method return files with extension.
	
	Parameters
	----------
	folder : path containing the files.
	ext : selected extensions.

	Return
	------
	ext_files : files in folder with the extensions in ext.
	"""
	# Basic check.
	if isinstance(ext, str): ext = [ext]
	# List files and add selected ones.
	all_files, ext_files = os.listdir(folder), []
	for f in all_files:
		for e in ext:
			if f.endswith(e):
				ext_files.append(f)
				break
	# Return
	return ext_files

#-------------------------------------------------------------------------
#                           GENERIC READ METHOD
#-------------------------------------------------------------------------
def read_data(ftype, **kwargs):
	"""This method loads data from the specified format.

	Parameters
	----------
	ftype : type of files to be load. Currently supported (xls, xlsx, csv, mysql).
	rename_map : dict-like, with current name and desired name.
	std_cols : list of columns that want to be kept.
	url : Only for 'mysql' - url to connect to the mysql sever.
	query : Only for 'mysql' - query to be executed against the server.

	Return
	------
	dataframe - <DataFrame> from pandas with content.
	"""
	if kwargs is None: return None
	if (ftype=='xls' or ftype=='xlsx'): return read_excel(**kwargs)
	if (ftype=='csv'):                  return read_csv(**kwargs)
	elif (ftype=='mysql'):              return read_mysql_server(**kwargs)
	else:                  		     	return None


#-------------------------------------------------------------------------
#                            READ MYSQL
#-------------------------------------------------------------------------
def read_mysql_server(url=None, query=None, rename_map=None):
	"""This function rename codes (i.e. PAER1,PAER2 to PEAR)

	Parameters
	----------

	Return
	------
	"""
	# Import library.
	from sqlalchemy import create_engine
	# Create connection.
	engine = create_engine(url)
	connection = engine.raw_connection()
	# Read SQL into DataFrame.
	df = pd.read_sql(query, connection)
	# Rename columns.
	df.rename(columns=rename_map, inplace=True)
	# Return
	return df


#-------------------------------------------------------------------------
#                            READ EXCEL
#-------------------------------------------------------------------------
def read_excel(path):
	"""This function...

	Parameters
	----------

	Return
	------
	"""
	if not os.path.exists(path): return None
	if os.path.isfile(path): return read_excel_file(path, **kwargs) 
	if os.path.isdir(path): return read_excel_folder(path, **kwargs)
	return None

def read_excel_folder(path=None, **kwargs):
	"""Load all excel files into a pandas dataframe.

	Parameters
	----------
	path : 

	Returns
	-------
	"""
	# Find files.
	files = files_with_extension(path, ['.xls','.xlsx'])
	if len(files)==0: return None
	print("Loading files...%s" % files)
	# Load data.
	df = None
	for fname in files:
		filepath = "%s/%s" % (path, fname)
		df_aux = read_excel_file(filepath, **kwargs)
		df = pd.concat([df, df_aux])
	# Remove duplicates.
	df = df.drop_duplicates()
	# Return
	return df

def read_excel_file(path, rename_map={}, std_cols=None, **kwargs):
	"""This method...

	Parameters
	----------

	Return
	------

	"""
	# Read data and format it.
	df = pd.read_excel(path, **kwargs) #, quotechar='"', quoting=1)
	df.rename(columns=rename_map, inplace=True)
	if std_cols is not None:
		df = pd.DataFrame(df, columns=std_cols)
	# Return
	return df



#-------------------------------------------------------------------------
#                            READ CSV
#-------------------------------------------------------------------------
def read_csv(path, **kwargs):
	"""This method...

	Parameters
	----------

	Return
	------
	"""
	if not os.path.exists(path): return None
	if os.path.isfile(path): return read_csv_file(path, **kwargs) 
	if os.path.isdir(path): return read_csv_folder(path, **kwargs)
	return None

def read_csv_folder(path, **kwargs):
	"""Load all excel files into a pandas dataframe.

	Parameters
	----------
	folder : 
	rename_map :
	std_cols :

	Returns
	-------
	"""
	# Find files.
	files = files_with_extension(path, '.csv')
	if len(files)==0: return None
	print ("Loading files...%s" % files)
	# Load data.
	df = None
	dfs = []
	for fname in files:
		filepath = "%s/%s" % (path, fname)
		df_aux = read_csv_file(filepath, **kwargs)
		df = pd.concat([df, df_aux])
	# Remove duplicates.
	df = df.drop_duplicates()
	# Return
	return df

def read_csv_file(path, rename_map={}, 
												std_cols=None, 
												keep_cols=None,
												date_cols=[],
												**kwargs):
	"""This method reads a csv file.

	Note: 1. It uses the pandas function pd.read_csv.
	      2.

	Note: keep_cols is used instead of default usecols from pandas because when
	reading different files we cannot ensure that all have the same column names
	before renaming.

	Parameters
	----------
	rename_map :
	std_cols   :
	keep_cols  :
	date_cols  :

	Return
	------

	"""
	# Read data and format it.
	df = pd.read_csv(path, **kwargs)
	# Rename columns.
	df.rename(columns=rename_map, inplace=True)
	# Parse dates.
	for c in date_cols:
		if c in df.columns:
				df[c] = pd.to_datetime(df[c], infer_datetime_format=True)
	# Columns that compose the dataframe.
	if std_cols is not None:
		df = pd.DataFrame(df, columns=std_cols)
	# Select some columns (useful if MemoryError)
	if keep_cols is not None:
		df = df[keep_cols]
	# Return
	return df


###############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
###############################################################################
# Libraries
import math
import pickle
import inspect
import warnings
import numpy as np
import pandas as pd
# import cPickle as pickle # not needed in python 3.x

# Specific
from copy import deepcopy
from sklearn.model_selection import ParameterGrid


# -----------------------------------------------------------------------------
#                              helper methods
# -----------------------------------------------------------------------------
def fargs(function, kwargs):
    """Finds parameters in kwargs tho use in function.

    Parameters
    ----------
    function : callable
      The function

    kwargs : dict-like
      The parameters and corresponding values.

    Returns
    -------
    """
    # Get all parameters for the function.
    prms = inspect.getfullargspec(function)
    # Create dictionary and return
    return {k: kwargs[k] for k in prms.args if k in kwargs}


def getargspecdict(instance, funcname):
    """This method creates a dictionary with pairs name and value.

    Parameters
    ----------
    instance : object with values
    funcname : function which parameters name will be looked for.

    Returns
    -------
    tpls : dictionary with argument name and value.
    """
    try:
        # Get argument parameters.
        func = getattr(instance, funcname, None)
        prms = inspect.getfullargspec(func)
        tpls = {}
        # Create and fill dictionary
        for name in prms.args:
            if name == 'self': continue
            tpls[name] = getattr(instance, name, None)
        # Return
        return tpls
    except Exception as e:
        # Print
        print("[Exception at _getargspecdict : %s" % e)
        # Return
        return {}


# def attrs(self):
#  """This method returns all the defined attributes as tuples."""
#  return inspect.getmembers(self, lambda a: not inspect.isroutine(a))

# def methods(self):
#  """This method returns all the defined methods as tuples."""
#  return inspect.getmembers(self, lambda a: inspect.isroutine(a))


class BaseWrapper(object):
    """Description



    """

    # Main attributes of the class.
    _raw = None  # The raw object (statsmodels, scipy, ...)
    _result = {}  # Dictionary with metrics filled in self._init_result().
    _config = {}  # Dictioanry with configuration filled in
    _conkwargs = {}
    _fitkwargs = {}

    def __init__(self, estimator=None, evaluate=True):
        """Constructor empty defined just so grid_search works in main.

        Note: Emptying the configuration attribute is important because when
              calling grid_search, the self.__class__(args) calls the __init__
              method of the instance and updates the self._config dictionary.
              Since it might not have been deepcopied, such modification will
              alter previous wrappers created during the grid search.
        """
        # Store the estimator
        self.estimator = estimator

        # Set the name
        self._name = self.__class__.__name__.replace('Wrapper', '')

        # Initialize the containers
        self._raw = None
        self._result = {}
        self._config = {}

        # ---------------------------------------------------------------------------

    #                            helper methods
    # ---------------------------------------------------------------------------
    def _empty(self):
        """This method empties the class."""
        self._raw = None
        self._result = {}
        self._config = {}

    def _identifier(self):
        """The name to identify this object."""
        return "%s" % self._name

    def __getattr__(self, name):
        """This method allows to retrieve series attributes with dot notation.

        .. note::

          The __getattr__ method is called only when __getattribute__ method was
          unsuccessful. As such, attributes such as .__class__ can still be
          called even though this method is overriden.

        Parameters
        ----------
        name : string
          The name of the attribute to retrieve.

        """
        # Retrieve the attribute
        if name in self._result: return self._result[name]
        if name in self._config: return self._config[name]
        # Raise error
        raise AttributeError("'%s' object has no attribute '%s'" % \
                             (self.__class__.__name__, name))

    # ---------------------------------------------------------------------------
    #                               save and load
    # ---------------------------------------------------------------------------
    def save(self, fname):
        """This method saves the wrapper."""
        pickle.dump(self.__dict__, open(fname, "wb"))

    def load(self, fname):
        """This method loads the wrapper."""
        self.__dict__.clear()
        self.__dict__.update(pickle.load(open(fname, "rb")))
        return self

        # ---------------------------------------------------------------------------

    #                               grid search
    # ---------------------------------------------------------------------------
    def grid_search(self, grid_params, verbose=0):
        """This method computes grid search.

        .. note: The wrapper needs to have the method ``fit`` implemented.

        Parameters
        ----------
        grid_params : array-like
          The grid of parameters to search.

        verbose : int
          Wether to show the progress of the search.

        Returns
        -------
        summary : summary with all the elements.
        """

        # Create empty list.
        wrappers = []

        # Create all possible combinations
        parameters = ParameterGrid(grid_params)

        # Number of combinations
        n = len(parameters)

        # Loop for all possible combinations.
        for i, params in enumerate(parameters):
            try:
                # Deep copy the object.
                copied = deepcopy(self)
                # Create model and fit
                wrappers.append(copied.fit(**params))
                # Show information
                if (verbose > 0):
                    print("%d/%d. %s" % (i + 1, n, copied._identifier()))

            except Exception as e:
                # Create message and raise warning
                msg = "%d/%d. %s ... failed: %s" % (i + 1, n, copied._identifier(), e)
                print(msg)

        # Return summary.
        return wrappers

    def from_list_dataframe(self, wrapper_list, **kwargs):
        """This methods creates a dataframe summary from a list.

        Parameters
        ----------
        wrapper_list : array-like
          The list with wrapper objects.

        flabel : boolean
          Wether to include a label before the attributes.

        label : string
          The label to include before the attributes. By default it includes
          the value of the attribute `self._name` in lowercase.

        Returns
        -------
        summary : pandas dataframe.
        """
        # Create summary.
        summary = pd.DataFrame()

        # Loop filling the summary.
        for i, wrapper in enumerate(wrapper_list):
            results = wrapper.as_series(**kwargs).rename(i)
            summary = pd.concat([summary, results], axis=1, join='outer')

        # Return
        return summary.T

    # ---------------------------------------------------------------------------
    #                                OVERRIDE
    # ---------------------------------------------------------------------------
    def as_series(self, flabel=True, label=None):
        """This method returns a series with all the information.

        Parameters
        ----------
        flabel : boolean
          Wether to include a label before the attributes.

        label : string
          The label to include before the attributes. By default it includes
          the value of the attribute `self._name` in lowercase.

        Returns
        -------
        series : pandas series with the results, configuration and model.
        """
        # Concatenate the configuration.
        s = {}
        s.update(self._result)
        s.update(self._config)
        s.update({'model': self._raw})
        s.update({'id': self._identifier()})

        # No label.
        if not flabel: return pd.Series(s)

        # Create the label to include
        if label is None and hasattr(self, '_name'):
            label = self._name.lower()
        else:
            label = str(label)

        # format label
        f = lambda x: "%s-%s" % (label.lower(), x)

        # Return
        return pd.Series(s).rename(index=f, copy=True)

    def as_summary(self):
        """This method displays the final summary."""
        # Call summary function from _raw (if exists).
        fsummary = getattr(self._raw, "summary", None)
        if callable(fsummary):
            return fsummary(self._raw)
        # Show the series information.
        return self.as_series().__repr__()

    # ---------------------------------------------------------------------------
    #                           method to override
    # ---------------------------------------------------------------------------
    def evaluate(self, **kwargs):
        """
        """
        # Return empty.
        return {}

    # ---------------------------------------------------------------------------
    #                                fit methods
    # ---------------------------------------------------------------------------
    def _fit_funct(self, **kwargs):
        """
        """
        # Get arguments
        arguments = fargs(self.estimator, kwargs)
        # Call function
        self._raw = self.estimator(**arguments)

    def _fit_class(self, **kwargs):
        """
        """
        # Get arguments
        conkwargs = fargs(self.estimator.__init__, kwargs)
        fitkwargs = fargs(self.estimator.fit, kwargs)
        # Call function
        self._raw = self.estimator(**conkwargs).fit(**fitkwargs)

    def fit(self, **kwargs):
        """This method performs the fit.

        Parameters
        ----------
        kwargs : dict-like
          The arguments that will be passed to the method.

        Returns
        -------

        """
        # Empty the class
        self._empty()
        # Update the configuration
        self._config.update(kwargs)

        # Fit the model
        if inspect.isfunction(self.estimator):
            self._fit_funct(**kwargs)
        elif inspect.isclass(self.estimator):
            self._fit_class(**kwargs)

        # Evaluate the model
        if self.evaluate:
            self._result = self.evaluate()

        # Return
        return self


if __name__ == '__main__':

    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 80)
    pd.set_option('display.precision', 4)

    # Create and fill a base statistic wrapper.
    w = BaseWrapper()
    w._raw = object()
    w._config = {'p': 2, 'c': 4}
    w._result = {'score': 25}

    # Get attributes tuple list.
    #print(w.attrs())

    # Get methods tuple list.
    #print(w.methods())

    # Get series with parameters.
    #print(w.as_series())

    # Print summary.
    #print(w.as_summary())

    # Quick access to an attribute.
    #print(w.score)

    # -----------
    # Grid search
    # -----------
    # Parameters
    con_params = {
        'con_1': [True],
        'con_2': ['1', '2'],
    }
    fit_params = {
        'fit_1': [5]
    }

    # Grid search method
    #summary = w.grid_search_dataframe(con_kwargs=con_params,
    #                                  fit_kwargs=fit_params)

    # Show
    #print(summary)



# Libraries generic
import pandas as pd
from pathlib import Path

def quickimport(path, connection):
    """


    :param path:
    :param connection:
    :return:
    """
    # -----------------------------------------------
    # Import lookup tables
    # -----------------------------------------------
    # Information
    print("Importing lookup tables...")

    # Lookup tables
    lookup = ['sensitivity.csv',
              'method.csv',
              'specimen.csv',
              'antimicrobial.csv',
              'microorganism.csv']

    keep = ['id', 'date_created', 'date_updated',
            'name', 'code', 'description', 'is_visible']

    for lp in lookup:
        # Read DataFrame
        df = pd.read_csv('%s/%s' % (path, lp))

        # Rename columns (remove in the future)
        #df.columns = [c.split("_")[-1] for c in df.columns]

        # Add dates
        df['date_created'] = pd.to_datetime('today')
        df['date_updated'] = pd.to_datetime('today')

        # Add missing
        if not 'description' in df:
            df['description'] = None
        if not 'is_visible' in df:
            df['is_visible'] = True

        # Fill na with '' so null=False.
        df = df.fillna(df.dtypes.replace({'O': ''}))

        # Keep
        df = df[keep]

        # Keep columns
        #print(df.name.value_counts())

        # Save
        try:
            df.to_sql(name='microbiology_%s' % Path(lp).stem,
                  con=connection,
                  #if_exists='replace',
                  if_exists='append',
                  index=False)
        except Exception as e:
            print(e)


    # ------------------------------------------------
    # Import patient
    # ------------------------------------------------
    df = pd.read_csv('%s/%s' % (path, 'patient.csv'))
    # Add dates
    df['date_created'] = pd.to_datetime('today')
    df['date_updated'] = pd.to_datetime('today')
    # Fill na with '' so null=False.
    df = df.fillna(df.dtypes.replace({'O': ''}))
    # Keep
    df = df.rename(columns={'name': 'hos_number'})
    df['name'] = ''
    df['surname'] = ''
    df['gender'] = 'M'
    df = df[['hos_number', 'id', 'date_created', 'date_updated', 'name',
             'surname', 'gender']]

    # Save
    try:
        df.to_sql(name='microbiology_%s' % 'patient',
                  con=connection,
                  # if_exists='replace',
                  chunksize=None,
                  if_exists='append',
                  index=False)
    except Exception as e:
        print(e)

    # ------------------------------------------------
    # Import data
    # ------------------------------------------------
    # Information
    print("Importing susceptibility tests...")

    # Load data
    data = pd.concat([pd.read_csv(f, #nrows=1000,
        encoding="ISO-8859-1", engine='c')
            for f in glob.glob(path + "/susceptibility-*.csv")])

    # Add missing
    data['date_created'] = pd.to_datetime('today')
    data['date_updated'] = pd.to_datetime('today')

    # Columns to keep
    keep = ['id',
            'date_created', 'date_updated', 'date_received', 'date_outcome',
            'laboratory_number',  'reported', 'antimicrobial_id',
            'method_id', 'microorganism_id', 'patient_id', 'sensitivity_id',
            'specimen_id'] #'mic',

    data = data[keep]

    # Save
    data.to_sql(name='microbiology_susceptibilitytest',
               con=connection,
               if_exists='append',
               index=False)




if __name__ == '__main__':

    # Libraries generic
    import glob
    import pandas as pd

    # Libraries specific
    from sqlalchemy import create_engine

    # Path
    path = './20210423-132035'

    # Connection parameters
    user = 'root'
    pwd = 'toor'
    host = 'localhost'
    schema = 'epicimpoc'
    fmt = 'mysql+pymysql://{user}:{pwd}@{host}/{schema}'

    # Create connection
    connection = create_engine(fmt.format(user=user,
        pwd=pwd, host=host, schema=schema))

    # Run quick import
    quickimport(path=path, connection=connection)


# Libraries
import os
import pytest
import pathlib
import runpy

from mock import patch

# .. note: When using <runpy>, the patch works so that we avoid
#          running the pyplot.show method but the code run does
#          not count for coverage. When using <os> the opposite
#          happens.

# Find the examples folder
examples = pathlib.Path(__file__, '../../../', 'examples').resolve()
pyamr_core = pathlib.Path(__file__, '../../', 'core').resolve()
pyamr_stats = pathlib.Path(__file__, '../../', 'core/stats').resolve()

# Find all the scripts
scripts_tutorial = (examples / 'tutorial').glob('**/*.py')
scripts_visualisation = (examples / 'visualisation').glob('**/*.py')
scripts_stats = list((pyamr_core / 'stats').glob('**/*.py'))
scripts_core = (pyamr_core / 'stats').glob('*.py')
scripts_metrics = (pyamr_core / 'stats').glob('*.py')

pyamr = pathlib.Path(__file__, '../../').resolve()

# --------------------------------------
# Run scripts
# --------------------------------------
"""
@patch('matplotlib.pyplot.show')
def test_script_run_pyamr(self):
    for f in (pyamr).glob('**/*.py'):
        runpy.run_path(str(f))
"""


@patch('matplotlib.pyplot.show')
def test_script_run_pyamr_metrics(self):
    for f in (pyamr / 'metrics').glob('**/*.py'):
        os.system('{} {}'.format('python', str(f)))
        #runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_pyamr_graphics(self):
    for f in (pyamr / 'graphics').glob('**/*.py'):
        os.system('{} {}'.format('python', str(f)))
        #runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_pyamr_core(self):
    for f in (pyamr / 'core').glob('*.py'):
        os.system('{} {}'.format('python', str(f)))
        #runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_pyamr_core_regression(self):
    for f in (pyamr / 'core' / 'regression').glob('**/*.py'):
        os.system('{} {}'.format('python', str(f)))
        #runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_pyamr_core_stats(self):
    for f in (pyamr / 'core' / 'stats').glob('**/*.py'):
        os.system('{} {}'.format('python', str(f)))
        #runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_pyamr_core_table(self):
    for f in (pyamr / 'core' / 'table').glob('**/*.py'):
        os.system('{} {}'.format('python', str(f)))
        #runpy.run_path(str(f))


"""
@patch('matplotlib.pyplot.show')
def test_script_run_core(self):
    for f in [pyamr_core / 'sari.py',
              pyamr_core / 'asai.py',
              pyamr_core / 'sart.py',
              pyamr_core / 'mari.py']:
        runpy.run_path(str(f))
"""
"""
def test_script_run_stats(self):
    #for f in [pyamr_stats / 'adfuller.py',
    #          pyamr_stats / 'correlation.py',
    #          pyamr_stats / 'kendall.py',
    #          pyamr_stats / 'kpss.py',
    #          pyamr_stats / 'stationarity.py']:
    #    print(f)
    #    #runpy.run_path(str(f))
    #    assert 4 == 5
    aux = pathlib.Path(__file__, '../../', 'core/stats').resolve()
    print(aux)
    assert 4==5
"""

"""
def test_script_execution_core_stats():
    f = pathlib.Path(__file__, '../../', 'core/stats').resolve()
    print(f / 'adfuller.py')
    runpy.run_path(str(f / 'adfuller.py'))
"""

"""
@pytest.mark.parametrize('script', scripts_stats)
def test_script_execution_core_stats(script):
    runpy.run_path(str(script))

@pytest.mark.parametrize('script', scripts_core)
def test_script_execution_core(script):
    runpy.run_path(str(script))
"""
"""
@pytest.mark.parametrize('script', scripts_tutorial)
def test_script_execution_widgets(script):
    runpy.run_path(str(script))
"""

"""
@patch('matplotlib.pyplot.show')
def test_script_run_visualisation_no_show(self):
    for f in scripts_visualisation:
        runpy.run_path(str(f))
"""

"""
@patch('matplotlib.pyplot.show')
def test_script_run_examples_no_show(self):
    for f in examples.glob('**/*.py'):
        runpy.run_path(str(f))
"""
"""
@patch('matplotlib.pyplot.show')
def test_script_run_examples_tutorials_no_show(self):
    for f in (examples / 'tutorials').glob('**/*.py'):
        runpy.run_path(str(f))
"""

"""
# .. note: For some reason, I have not managed to combine
#          parametrize and mock together into one single
#          method.
@patch('matplotlib.pyplot.show')
def test_script_run_examples_tutorials_no_show(self):
    for f in (examples / 'tutorials').glob('**/*.py'):
        runpy.run_path(str(f))
"""

"""
@patch('matplotlib.pyplot.show')
def test_script_run_examples_indexes_no_show(self):
    for f in (examples / 'indexes').glob('**/*.py'):
        runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_examples_forecasting_no_show(self):
    for f in (examples / 'forecasting').glob('**/*.py'):
        runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_examples_indexes_no_show(self):
    for f in (examples / 'indexes').glob('**/*.py'):
        runpy.run_path(str(f))

@patch('matplotlib.pyplot.show')
def test_script_run_examples_visualization_no_show(self):
    for f in (examples / 'visualization').glob('**/*.py'):
        runpy.run_path(str(f))
"""